(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global = typeof globalThis !== 'undefined' ? globalThis : global || self, global["axes-highlights"] = factory());
})(this, (function () { 'use strict';

	const DEFAULT_STATE = {
		x_enabled: false,

		x_lines: "",
		x_line_color: null,
		x_line_width: 2,
		x_line_dash: "2",

		x_areas: "",
		x_area_color: null,
		x_area_opacity: 0.1,
		x_area_texture: "none",
		x_font_size: 0.9,
		x_font_weight: "regular",
		x_label_vertical: true,
		x_label_color: null,
		x_label_max_width: 3,
		x_label_y_position: "top",
		x_label_x_position: "left",

		x_shadow_enabled: true,
		x_shadow_color: null,

		y_enabled: false,

		y_lines: "",
		y_line_color: null,
		y_line_width: 2,
		y_line_dash: "2",

		y_areas: "",
		y_area_color: null,
		y_area_opacity: 0.1,
		y_area_texture: "none",

		y_font_size: 0.9,
		y_font_weight: "regular",
		y_label_x_position: "right",
		y_label_y_position: "above",
		y_label_color: null,
		y_label_max_width: 3,

		y_shadow_enabled: true,
		y_shadow_color: null,
	};

	function initAnimationDuration(instance) {
		let animation_duration = 0;

		instance.animationDuration = function(value) {
			// If value is undefined we return the saved value
			if (value === undefined) return animation_duration;
			// If value is defined we set the saved value to the new value
			animation_duration = value;
			// Then return the instance to allow for method chaining
			return instance;
		};
	}

	function initNumberParser(instance) {
		let numberParser = (a) => { return a; };

		instance.numberParser = function(value) {
			if (value === undefined) return numberParser;
			numberParser = value;
			return instance;
		};
	}

	function initChartLayout(instance) {
		let chart_layout;

		instance.chartLayout = function(value) {
			if (value === undefined) return chart_layout;
			chart_layout = value;
			return instance;
		};
	}

	function initXDisabled(instance) {
		let x_disabled = false;

		instance.xDisabled = function(value) {
			if (value === undefined) return x_disabled;
			x_disabled = value;
			return instance;
		};
	}

	function initYDisabled(instance) {
		let y_disabled = false;

		instance.yDisabled = function(value) {
			if (value === undefined) return y_disabled;
			y_disabled = value;
			return instance;
		};
	}

	function initXExtendByStep(instance) {
		let x_extend_by_step = false;

		instance.xExtendByStep = function(value) {
			if (value === undefined) return x_extend_by_step;
			x_extend_by_step = value;
			return instance;
		};
	}

	function initYExtendByStep(instance) {
		let y_extend_by_step = false;

		instance.yExtendByStep = function(value) {
			if (value === undefined) return y_extend_by_step;
			y_extend_by_step = value;
			return instance;
		};
	}

	var xhtml = "http://www.w3.org/1999/xhtml";

	var namespaces = {
	  svg: "http://www.w3.org/2000/svg",
	  xhtml: xhtml,
	  xlink: "http://www.w3.org/1999/xlink",
	  xml: "http://www.w3.org/XML/1998/namespace",
	  xmlns: "http://www.w3.org/2000/xmlns/"
	};

	function namespace(name) {
	  var prefix = name += "", i = prefix.indexOf(":");
	  if (i >= 0 && (prefix = name.slice(0, i)) !== "xmlns") name = name.slice(i + 1);
	  return namespaces.hasOwnProperty(prefix) ? {space: namespaces[prefix], local: name} : name; // eslint-disable-line no-prototype-builtins
	}

	function creatorInherit(name) {
	  return function() {
	    var document = this.ownerDocument,
	        uri = this.namespaceURI;
	    return uri === xhtml && document.documentElement.namespaceURI === xhtml
	        ? document.createElement(name)
	        : document.createElementNS(uri, name);
	  };
	}

	function creatorFixed(fullname) {
	  return function() {
	    return this.ownerDocument.createElementNS(fullname.space, fullname.local);
	  };
	}

	function creator(name) {
	  var fullname = namespace(name);
	  return (fullname.local
	      ? creatorFixed
	      : creatorInherit)(fullname);
	}

	function none() {}

	function selector(selector) {
	  return selector == null ? none : function() {
	    return this.querySelector(selector);
	  };
	}

	function selection_select(select) {
	  if (typeof select !== "function") select = selector(select);

	  for (var groups = this._groups, m = groups.length, subgroups = new Array(m), j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, subgroup = subgroups[j] = new Array(n), node, subnode, i = 0; i < n; ++i) {
	      if ((node = group[i]) && (subnode = select.call(node, node.__data__, i, group))) {
	        if ("__data__" in node) subnode.__data__ = node.__data__;
	        subgroup[i] = subnode;
	      }
	    }
	  }

	  return new Selection$1(subgroups, this._parents);
	}

	function array(x) {
	  return typeof x === "object" && "length" in x
	    ? x // Array, TypedArray, NodeList, array-like
	    : Array.from(x); // Map, Set, iterable, string, or anything else
	}

	function empty() {
	  return [];
	}

	function selectorAll(selector) {
	  return selector == null ? empty : function() {
	    return this.querySelectorAll(selector);
	  };
	}

	function arrayAll(select) {
	  return function() {
	    var group = select.apply(this, arguments);
	    return group == null ? [] : array(group);
	  };
	}

	function selection_selectAll(select) {
	  if (typeof select === "function") select = arrayAll(select);
	  else select = selectorAll(select);

	  for (var groups = this._groups, m = groups.length, subgroups = [], parents = [], j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, node, i = 0; i < n; ++i) {
	      if (node = group[i]) {
	        subgroups.push(select.call(node, node.__data__, i, group));
	        parents.push(node);
	      }
	    }
	  }

	  return new Selection$1(subgroups, parents);
	}

	function matcher(selector) {
	  return function() {
	    return this.matches(selector);
	  };
	}

	function childMatcher(selector) {
	  return function(node) {
	    return node.matches(selector);
	  };
	}

	var find = Array.prototype.find;

	function childFind(match) {
	  return function() {
	    return find.call(this.children, match);
	  };
	}

	function childFirst() {
	  return this.firstElementChild;
	}

	function selection_selectChild(match) {
	  return this.select(match == null ? childFirst
	      : childFind(typeof match === "function" ? match : childMatcher(match)));
	}

	var filter = Array.prototype.filter;

	function children() {
	  return this.children;
	}

	function childrenFilter(match) {
	  return function() {
	    return filter.call(this.children, match);
	  };
	}

	function selection_selectChildren(match) {
	  return this.selectAll(match == null ? children
	      : childrenFilter(typeof match === "function" ? match : childMatcher(match)));
	}

	function selection_filter(match) {
	  if (typeof match !== "function") match = matcher(match);

	  for (var groups = this._groups, m = groups.length, subgroups = new Array(m), j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, subgroup = subgroups[j] = [], node, i = 0; i < n; ++i) {
	      if ((node = group[i]) && match.call(node, node.__data__, i, group)) {
	        subgroup.push(node);
	      }
	    }
	  }

	  return new Selection$1(subgroups, this._parents);
	}

	function sparse(update) {
	  return new Array(update.length);
	}

	function selection_enter() {
	  return new Selection$1(this._enter || this._groups.map(sparse), this._parents);
	}

	function EnterNode(parent, datum) {
	  this.ownerDocument = parent.ownerDocument;
	  this.namespaceURI = parent.namespaceURI;
	  this._next = null;
	  this._parent = parent;
	  this.__data__ = datum;
	}

	EnterNode.prototype = {
	  constructor: EnterNode,
	  appendChild: function(child) { return this._parent.insertBefore(child, this._next); },
	  insertBefore: function(child, next) { return this._parent.insertBefore(child, next); },
	  querySelector: function(selector) { return this._parent.querySelector(selector); },
	  querySelectorAll: function(selector) { return this._parent.querySelectorAll(selector); }
	};

	function constant$1(x) {
	  return function() {
	    return x;
	  };
	}

	function bindIndex(parent, group, enter, update, exit, data) {
	  var i = 0,
	      node,
	      groupLength = group.length,
	      dataLength = data.length;

	  // Put any non-null nodes that fit into update.
	  // Put any null nodes into enter.
	  // Put any remaining data into enter.
	  for (; i < dataLength; ++i) {
	    if (node = group[i]) {
	      node.__data__ = data[i];
	      update[i] = node;
	    } else {
	      enter[i] = new EnterNode(parent, data[i]);
	    }
	  }

	  // Put any non-null nodes that don’t fit into exit.
	  for (; i < groupLength; ++i) {
	    if (node = group[i]) {
	      exit[i] = node;
	    }
	  }
	}

	function bindKey(parent, group, enter, update, exit, data, key) {
	  var i,
	      node,
	      nodeByKeyValue = new Map,
	      groupLength = group.length,
	      dataLength = data.length,
	      keyValues = new Array(groupLength),
	      keyValue;

	  // Compute the key for each node.
	  // If multiple nodes have the same key, the duplicates are added to exit.
	  for (i = 0; i < groupLength; ++i) {
	    if (node = group[i]) {
	      keyValues[i] = keyValue = key.call(node, node.__data__, i, group) + "";
	      if (nodeByKeyValue.has(keyValue)) {
	        exit[i] = node;
	      } else {
	        nodeByKeyValue.set(keyValue, node);
	      }
	    }
	  }

	  // Compute the key for each datum.
	  // If there a node associated with this key, join and add it to update.
	  // If there is not (or the key is a duplicate), add it to enter.
	  for (i = 0; i < dataLength; ++i) {
	    keyValue = key.call(parent, data[i], i, data) + "";
	    if (node = nodeByKeyValue.get(keyValue)) {
	      update[i] = node;
	      node.__data__ = data[i];
	      nodeByKeyValue.delete(keyValue);
	    } else {
	      enter[i] = new EnterNode(parent, data[i]);
	    }
	  }

	  // Add any remaining nodes that were not bound to data to exit.
	  for (i = 0; i < groupLength; ++i) {
	    if ((node = group[i]) && (nodeByKeyValue.get(keyValues[i]) === node)) {
	      exit[i] = node;
	    }
	  }
	}

	function datum(node) {
	  return node.__data__;
	}

	function selection_data(value, key) {
	  if (!arguments.length) return Array.from(this, datum);

	  var bind = key ? bindKey : bindIndex,
	      parents = this._parents,
	      groups = this._groups;

	  if (typeof value !== "function") value = constant$1(value);

	  for (var m = groups.length, update = new Array(m), enter = new Array(m), exit = new Array(m), j = 0; j < m; ++j) {
	    var parent = parents[j],
	        group = groups[j],
	        groupLength = group.length,
	        data = array(value.call(parent, parent && parent.__data__, j, parents)),
	        dataLength = data.length,
	        enterGroup = enter[j] = new Array(dataLength),
	        updateGroup = update[j] = new Array(dataLength),
	        exitGroup = exit[j] = new Array(groupLength);

	    bind(parent, group, enterGroup, updateGroup, exitGroup, data, key);

	    // Now connect the enter nodes to their following update node, such that
	    // appendChild can insert the materialized enter node before this node,
	    // rather than at the end of the parent node.
	    for (var i0 = 0, i1 = 0, previous, next; i0 < dataLength; ++i0) {
	      if (previous = enterGroup[i0]) {
	        if (i0 >= i1) i1 = i0 + 1;
	        while (!(next = updateGroup[i1]) && ++i1 < dataLength);
	        previous._next = next || null;
	      }
	    }
	  }

	  update = new Selection$1(update, parents);
	  update._enter = enter;
	  update._exit = exit;
	  return update;
	}

	function selection_exit() {
	  return new Selection$1(this._exit || this._groups.map(sparse), this._parents);
	}

	function selection_join(onenter, onupdate, onexit) {
	  var enter = this.enter(), update = this, exit = this.exit();
	  enter = typeof onenter === "function" ? onenter(enter) : enter.append(onenter + "");
	  if (onupdate != null) update = onupdate(update);
	  if (onexit == null) exit.remove(); else onexit(exit);
	  return enter && update ? enter.merge(update).order() : update;
	}

	function selection_merge(selection) {
	  if (!(selection instanceof Selection$1)) throw new Error("invalid merge");

	  for (var groups0 = this._groups, groups1 = selection._groups, m0 = groups0.length, m1 = groups1.length, m = Math.min(m0, m1), merges = new Array(m0), j = 0; j < m; ++j) {
	    for (var group0 = groups0[j], group1 = groups1[j], n = group0.length, merge = merges[j] = new Array(n), node, i = 0; i < n; ++i) {
	      if (node = group0[i] || group1[i]) {
	        merge[i] = node;
	      }
	    }
	  }

	  for (; j < m0; ++j) {
	    merges[j] = groups0[j];
	  }

	  return new Selection$1(merges, this._parents);
	}

	function selection_order() {

	  for (var groups = this._groups, j = -1, m = groups.length; ++j < m;) {
	    for (var group = groups[j], i = group.length - 1, next = group[i], node; --i >= 0;) {
	      if (node = group[i]) {
	        if (next && node.compareDocumentPosition(next) ^ 4) next.parentNode.insertBefore(node, next);
	        next = node;
	      }
	    }
	  }

	  return this;
	}

	function selection_sort(compare) {
	  if (!compare) compare = ascending;

	  function compareNode(a, b) {
	    return a && b ? compare(a.__data__, b.__data__) : !a - !b;
	  }

	  for (var groups = this._groups, m = groups.length, sortgroups = new Array(m), j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, sortgroup = sortgroups[j] = new Array(n), node, i = 0; i < n; ++i) {
	      if (node = group[i]) {
	        sortgroup[i] = node;
	      }
	    }
	    sortgroup.sort(compareNode);
	  }

	  return new Selection$1(sortgroups, this._parents).order();
	}

	function ascending(a, b) {
	  return a < b ? -1 : a > b ? 1 : a >= b ? 0 : NaN;
	}

	function selection_call() {
	  var callback = arguments[0];
	  arguments[0] = this;
	  callback.apply(null, arguments);
	  return this;
	}

	function selection_nodes() {
	  return Array.from(this);
	}

	function selection_node() {

	  for (var groups = this._groups, j = 0, m = groups.length; j < m; ++j) {
	    for (var group = groups[j], i = 0, n = group.length; i < n; ++i) {
	      var node = group[i];
	      if (node) return node;
	    }
	  }

	  return null;
	}

	function selection_size() {
	  let size = 0;
	  for (const node of this) ++size; // eslint-disable-line no-unused-vars
	  return size;
	}

	function selection_empty() {
	  return !this.node();
	}

	function selection_each(callback) {

	  for (var groups = this._groups, j = 0, m = groups.length; j < m; ++j) {
	    for (var group = groups[j], i = 0, n = group.length, node; i < n; ++i) {
	      if (node = group[i]) callback.call(node, node.__data__, i, group);
	    }
	  }

	  return this;
	}

	function attrRemove$1(name) {
	  return function() {
	    this.removeAttribute(name);
	  };
	}

	function attrRemoveNS$1(fullname) {
	  return function() {
	    this.removeAttributeNS(fullname.space, fullname.local);
	  };
	}

	function attrConstant$1(name, value) {
	  return function() {
	    this.setAttribute(name, value);
	  };
	}

	function attrConstantNS$1(fullname, value) {
	  return function() {
	    this.setAttributeNS(fullname.space, fullname.local, value);
	  };
	}

	function attrFunction$1(name, value) {
	  return function() {
	    var v = value.apply(this, arguments);
	    if (v == null) this.removeAttribute(name);
	    else this.setAttribute(name, v);
	  };
	}

	function attrFunctionNS$1(fullname, value) {
	  return function() {
	    var v = value.apply(this, arguments);
	    if (v == null) this.removeAttributeNS(fullname.space, fullname.local);
	    else this.setAttributeNS(fullname.space, fullname.local, v);
	  };
	}

	function selection_attr(name, value) {
	  var fullname = namespace(name);

	  if (arguments.length < 2) {
	    var node = this.node();
	    return fullname.local
	        ? node.getAttributeNS(fullname.space, fullname.local)
	        : node.getAttribute(fullname);
	  }

	  return this.each((value == null
	      ? (fullname.local ? attrRemoveNS$1 : attrRemove$1) : (typeof value === "function"
	      ? (fullname.local ? attrFunctionNS$1 : attrFunction$1)
	      : (fullname.local ? attrConstantNS$1 : attrConstant$1)))(fullname, value));
	}

	function defaultView(node) {
	  return (node.ownerDocument && node.ownerDocument.defaultView) // node is a Node
	      || (node.document && node) // node is a Window
	      || node.defaultView; // node is a Document
	}

	function styleRemove$1(name) {
	  return function() {
	    this.style.removeProperty(name);
	  };
	}

	function styleConstant$1(name, value, priority) {
	  return function() {
	    this.style.setProperty(name, value, priority);
	  };
	}

	function styleFunction$1(name, value, priority) {
	  return function() {
	    var v = value.apply(this, arguments);
	    if (v == null) this.style.removeProperty(name);
	    else this.style.setProperty(name, v, priority);
	  };
	}

	function selection_style(name, value, priority) {
	  return arguments.length > 1
	      ? this.each((value == null
	            ? styleRemove$1 : typeof value === "function"
	            ? styleFunction$1
	            : styleConstant$1)(name, value, priority == null ? "" : priority))
	      : styleValue(this.node(), name);
	}

	function styleValue(node, name) {
	  return node.style.getPropertyValue(name)
	      || defaultView(node).getComputedStyle(node, null).getPropertyValue(name);
	}

	function propertyRemove(name) {
	  return function() {
	    delete this[name];
	  };
	}

	function propertyConstant(name, value) {
	  return function() {
	    this[name] = value;
	  };
	}

	function propertyFunction(name, value) {
	  return function() {
	    var v = value.apply(this, arguments);
	    if (v == null) delete this[name];
	    else this[name] = v;
	  };
	}

	function selection_property(name, value) {
	  return arguments.length > 1
	      ? this.each((value == null
	          ? propertyRemove : typeof value === "function"
	          ? propertyFunction
	          : propertyConstant)(name, value))
	      : this.node()[name];
	}

	function classArray(string) {
	  return string.trim().split(/^|\s+/);
	}

	function classList(node) {
	  return node.classList || new ClassList(node);
	}

	function ClassList(node) {
	  this._node = node;
	  this._names = classArray(node.getAttribute("class") || "");
	}

	ClassList.prototype = {
	  add: function(name) {
	    var i = this._names.indexOf(name);
	    if (i < 0) {
	      this._names.push(name);
	      this._node.setAttribute("class", this._names.join(" "));
	    }
	  },
	  remove: function(name) {
	    var i = this._names.indexOf(name);
	    if (i >= 0) {
	      this._names.splice(i, 1);
	      this._node.setAttribute("class", this._names.join(" "));
	    }
	  },
	  contains: function(name) {
	    return this._names.indexOf(name) >= 0;
	  }
	};

	function classedAdd(node, names) {
	  var list = classList(node), i = -1, n = names.length;
	  while (++i < n) list.add(names[i]);
	}

	function classedRemove(node, names) {
	  var list = classList(node), i = -1, n = names.length;
	  while (++i < n) list.remove(names[i]);
	}

	function classedTrue(names) {
	  return function() {
	    classedAdd(this, names);
	  };
	}

	function classedFalse(names) {
	  return function() {
	    classedRemove(this, names);
	  };
	}

	function classedFunction(names, value) {
	  return function() {
	    (value.apply(this, arguments) ? classedAdd : classedRemove)(this, names);
	  };
	}

	function selection_classed(name, value) {
	  var names = classArray(name + "");

	  if (arguments.length < 2) {
	    var list = classList(this.node()), i = -1, n = names.length;
	    while (++i < n) if (!list.contains(names[i])) return false;
	    return true;
	  }

	  return this.each((typeof value === "function"
	      ? classedFunction : value
	      ? classedTrue
	      : classedFalse)(names, value));
	}

	function textRemove() {
	  this.textContent = "";
	}

	function textConstant$1(value) {
	  return function() {
	    this.textContent = value;
	  };
	}

	function textFunction$1(value) {
	  return function() {
	    var v = value.apply(this, arguments);
	    this.textContent = v == null ? "" : v;
	  };
	}

	function selection_text(value) {
	  return arguments.length
	      ? this.each(value == null
	          ? textRemove : (typeof value === "function"
	          ? textFunction$1
	          : textConstant$1)(value))
	      : this.node().textContent;
	}

	function htmlRemove() {
	  this.innerHTML = "";
	}

	function htmlConstant(value) {
	  return function() {
	    this.innerHTML = value;
	  };
	}

	function htmlFunction(value) {
	  return function() {
	    var v = value.apply(this, arguments);
	    this.innerHTML = v == null ? "" : v;
	  };
	}

	function selection_html(value) {
	  return arguments.length
	      ? this.each(value == null
	          ? htmlRemove : (typeof value === "function"
	          ? htmlFunction
	          : htmlConstant)(value))
	      : this.node().innerHTML;
	}

	function raise() {
	  if (this.nextSibling) this.parentNode.appendChild(this);
	}

	function selection_raise() {
	  return this.each(raise);
	}

	function lower() {
	  if (this.previousSibling) this.parentNode.insertBefore(this, this.parentNode.firstChild);
	}

	function selection_lower() {
	  return this.each(lower);
	}

	function selection_append(name) {
	  var create = typeof name === "function" ? name : creator(name);
	  return this.select(function() {
	    return this.appendChild(create.apply(this, arguments));
	  });
	}

	function constantNull() {
	  return null;
	}

	function selection_insert(name, before) {
	  var create = typeof name === "function" ? name : creator(name),
	      select = before == null ? constantNull : typeof before === "function" ? before : selector(before);
	  return this.select(function() {
	    return this.insertBefore(create.apply(this, arguments), select.apply(this, arguments) || null);
	  });
	}

	function remove() {
	  var parent = this.parentNode;
	  if (parent) parent.removeChild(this);
	}

	function selection_remove() {
	  return this.each(remove);
	}

	function selection_cloneShallow() {
	  var clone = this.cloneNode(false), parent = this.parentNode;
	  return parent ? parent.insertBefore(clone, this.nextSibling) : clone;
	}

	function selection_cloneDeep() {
	  var clone = this.cloneNode(true), parent = this.parentNode;
	  return parent ? parent.insertBefore(clone, this.nextSibling) : clone;
	}

	function selection_clone(deep) {
	  return this.select(deep ? selection_cloneDeep : selection_cloneShallow);
	}

	function selection_datum(value) {
	  return arguments.length
	      ? this.property("__data__", value)
	      : this.node().__data__;
	}

	function contextListener(listener) {
	  return function(event) {
	    listener.call(this, event, this.__data__);
	  };
	}

	function parseTypenames$1(typenames) {
	  return typenames.trim().split(/^|\s+/).map(function(t) {
	    var name = "", i = t.indexOf(".");
	    if (i >= 0) name = t.slice(i + 1), t = t.slice(0, i);
	    return {type: t, name: name};
	  });
	}

	function onRemove(typename) {
	  return function() {
	    var on = this.__on;
	    if (!on) return;
	    for (var j = 0, i = -1, m = on.length, o; j < m; ++j) {
	      if (o = on[j], (!typename.type || o.type === typename.type) && o.name === typename.name) {
	        this.removeEventListener(o.type, o.listener, o.options);
	      } else {
	        on[++i] = o;
	      }
	    }
	    if (++i) on.length = i;
	    else delete this.__on;
	  };
	}

	function onAdd(typename, value, options) {
	  return function() {
	    var on = this.__on, o, listener = contextListener(value);
	    if (on) for (var j = 0, m = on.length; j < m; ++j) {
	      if ((o = on[j]).type === typename.type && o.name === typename.name) {
	        this.removeEventListener(o.type, o.listener, o.options);
	        this.addEventListener(o.type, o.listener = listener, o.options = options);
	        o.value = value;
	        return;
	      }
	    }
	    this.addEventListener(typename.type, listener, options);
	    o = {type: typename.type, name: typename.name, value: value, listener: listener, options: options};
	    if (!on) this.__on = [o];
	    else on.push(o);
	  };
	}

	function selection_on(typename, value, options) {
	  var typenames = parseTypenames$1(typename + ""), i, n = typenames.length, t;

	  if (arguments.length < 2) {
	    var on = this.node().__on;
	    if (on) for (var j = 0, m = on.length, o; j < m; ++j) {
	      for (i = 0, o = on[j]; i < n; ++i) {
	        if ((t = typenames[i]).type === o.type && t.name === o.name) {
	          return o.value;
	        }
	      }
	    }
	    return;
	  }

	  on = value ? onAdd : onRemove;
	  for (i = 0; i < n; ++i) this.each(on(typenames[i], value, options));
	  return this;
	}

	function dispatchEvent(node, type, params) {
	  var window = defaultView(node),
	      event = window.CustomEvent;

	  if (typeof event === "function") {
	    event = new event(type, params);
	  } else {
	    event = window.document.createEvent("Event");
	    if (params) event.initEvent(type, params.bubbles, params.cancelable), event.detail = params.detail;
	    else event.initEvent(type, false, false);
	  }

	  node.dispatchEvent(event);
	}

	function dispatchConstant(type, params) {
	  return function() {
	    return dispatchEvent(this, type, params);
	  };
	}

	function dispatchFunction(type, params) {
	  return function() {
	    return dispatchEvent(this, type, params.apply(this, arguments));
	  };
	}

	function selection_dispatch(type, params) {
	  return this.each((typeof params === "function"
	      ? dispatchFunction
	      : dispatchConstant)(type, params));
	}

	function* selection_iterator() {
	  for (var groups = this._groups, j = 0, m = groups.length; j < m; ++j) {
	    for (var group = groups[j], i = 0, n = group.length, node; i < n; ++i) {
	      if (node = group[i]) yield node;
	    }
	  }
	}

	var root = [null];

	function Selection$1(groups, parents) {
	  this._groups = groups;
	  this._parents = parents;
	}

	function selection() {
	  return new Selection$1([[document.documentElement]], root);
	}

	function selection_selection() {
	  return this;
	}

	Selection$1.prototype = selection.prototype = {
	  constructor: Selection$1,
	  select: selection_select,
	  selectAll: selection_selectAll,
	  selectChild: selection_selectChild,
	  selectChildren: selection_selectChildren,
	  filter: selection_filter,
	  data: selection_data,
	  enter: selection_enter,
	  exit: selection_exit,
	  join: selection_join,
	  merge: selection_merge,
	  selection: selection_selection,
	  order: selection_order,
	  sort: selection_sort,
	  call: selection_call,
	  nodes: selection_nodes,
	  node: selection_node,
	  size: selection_size,
	  empty: selection_empty,
	  each: selection_each,
	  attr: selection_attr,
	  style: selection_style,
	  property: selection_property,
	  classed: selection_classed,
	  text: selection_text,
	  html: selection_html,
	  raise: selection_raise,
	  lower: selection_lower,
	  append: selection_append,
	  insert: selection_insert,
	  remove: selection_remove,
	  clone: selection_clone,
	  datum: selection_datum,
	  on: selection_on,
	  dispatch: selection_dispatch,
	  [Symbol.iterator]: selection_iterator
	};

	function select(selector) {
	  return typeof selector === "string"
	      ? new Selection$1([[document.querySelector(selector)]], [document.documentElement])
	      : new Selection$1([[selector]], root);
	}

	function create$1(name) {
	  return select(creator(name).call(document.documentElement));
	}

	var noop = {value: function() {}};

	function dispatch() {
	  for (var i = 0, n = arguments.length, _ = {}, t; i < n; ++i) {
	    if (!(t = arguments[i] + "") || (t in _) || /[\s.]/.test(t)) throw new Error("illegal type: " + t);
	    _[t] = [];
	  }
	  return new Dispatch(_);
	}

	function Dispatch(_) {
	  this._ = _;
	}

	function parseTypenames(typenames, types) {
	  return typenames.trim().split(/^|\s+/).map(function(t) {
	    var name = "", i = t.indexOf(".");
	    if (i >= 0) name = t.slice(i + 1), t = t.slice(0, i);
	    if (t && !types.hasOwnProperty(t)) throw new Error("unknown type: " + t);
	    return {type: t, name: name};
	  });
	}

	Dispatch.prototype = dispatch.prototype = {
	  constructor: Dispatch,
	  on: function(typename, callback) {
	    var _ = this._,
	        T = parseTypenames(typename + "", _),
	        t,
	        i = -1,
	        n = T.length;

	    // If no callback was specified, return the callback of the given type and name.
	    if (arguments.length < 2) {
	      while (++i < n) if ((t = (typename = T[i]).type) && (t = get$1(_[t], typename.name))) return t;
	      return;
	    }

	    // If a type was specified, set the callback for the given type and name.
	    // Otherwise, if a null callback was specified, remove callbacks of the given name.
	    if (callback != null && typeof callback !== "function") throw new Error("invalid callback: " + callback);
	    while (++i < n) {
	      if (t = (typename = T[i]).type) _[t] = set$1(_[t], typename.name, callback);
	      else if (callback == null) for (t in _) _[t] = set$1(_[t], typename.name, null);
	    }

	    return this;
	  },
	  copy: function() {
	    var copy = {}, _ = this._;
	    for (var t in _) copy[t] = _[t].slice();
	    return new Dispatch(copy);
	  },
	  call: function(type, that) {
	    if ((n = arguments.length - 2) > 0) for (var args = new Array(n), i = 0, n, t; i < n; ++i) args[i] = arguments[i + 2];
	    if (!this._.hasOwnProperty(type)) throw new Error("unknown type: " + type);
	    for (t = this._[type], i = 0, n = t.length; i < n; ++i) t[i].value.apply(that, args);
	  },
	  apply: function(type, that, args) {
	    if (!this._.hasOwnProperty(type)) throw new Error("unknown type: " + type);
	    for (var t = this._[type], i = 0, n = t.length; i < n; ++i) t[i].value.apply(that, args);
	  }
	};

	function get$1(type, name) {
	  for (var i = 0, n = type.length, c; i < n; ++i) {
	    if ((c = type[i]).name === name) {
	      return c.value;
	    }
	  }
	}

	function set$1(type, name, callback) {
	  for (var i = 0, n = type.length; i < n; ++i) {
	    if (type[i].name === name) {
	      type[i] = noop, type = type.slice(0, i).concat(type.slice(i + 1));
	      break;
	    }
	  }
	  if (callback != null) type.push({name: name, value: callback});
	  return type;
	}

	var frame = 0, // is an animation frame pending?
	    timeout$1 = 0, // is a timeout pending?
	    interval = 0, // are any timers active?
	    pokeDelay = 1000, // how frequently we check for clock skew
	    taskHead,
	    taskTail,
	    clockLast = 0,
	    clockNow = 0,
	    clockSkew = 0,
	    clock = typeof performance === "object" && performance.now ? performance : Date,
	    setFrame = typeof window === "object" && window.requestAnimationFrame ? window.requestAnimationFrame.bind(window) : function(f) { setTimeout(f, 17); };

	function now() {
	  return clockNow || (setFrame(clearNow), clockNow = clock.now() + clockSkew);
	}

	function clearNow() {
	  clockNow = 0;
	}

	function Timer() {
	  this._call =
	  this._time =
	  this._next = null;
	}

	Timer.prototype = timer.prototype = {
	  constructor: Timer,
	  restart: function(callback, delay, time) {
	    if (typeof callback !== "function") throw new TypeError("callback is not a function");
	    time = (time == null ? now() : +time) + (delay == null ? 0 : +delay);
	    if (!this._next && taskTail !== this) {
	      if (taskTail) taskTail._next = this;
	      else taskHead = this;
	      taskTail = this;
	    }
	    this._call = callback;
	    this._time = time;
	    sleep();
	  },
	  stop: function() {
	    if (this._call) {
	      this._call = null;
	      this._time = Infinity;
	      sleep();
	    }
	  }
	};

	function timer(callback, delay, time) {
	  var t = new Timer;
	  t.restart(callback, delay, time);
	  return t;
	}

	function timerFlush() {
	  now(); // Get the current time, if not already set.
	  ++frame; // Pretend we’ve set an alarm, if we haven’t already.
	  var t = taskHead, e;
	  while (t) {
	    if ((e = clockNow - t._time) >= 0) t._call.call(null, e);
	    t = t._next;
	  }
	  --frame;
	}

	function wake() {
	  clockNow = (clockLast = clock.now()) + clockSkew;
	  frame = timeout$1 = 0;
	  try {
	    timerFlush();
	  } finally {
	    frame = 0;
	    nap();
	    clockNow = 0;
	  }
	}

	function poke() {
	  var now = clock.now(), delay = now - clockLast;
	  if (delay > pokeDelay) clockSkew -= delay, clockLast = now;
	}

	function nap() {
	  var t0, t1 = taskHead, t2, time = Infinity;
	  while (t1) {
	    if (t1._call) {
	      if (time > t1._time) time = t1._time;
	      t0 = t1, t1 = t1._next;
	    } else {
	      t2 = t1._next, t1._next = null;
	      t1 = t0 ? t0._next = t2 : taskHead = t2;
	    }
	  }
	  taskTail = t0;
	  sleep(time);
	}

	function sleep(time) {
	  if (frame) return; // Soonest alarm already set, or will be.
	  if (timeout$1) timeout$1 = clearTimeout(timeout$1);
	  var delay = time - clockNow; // Strictly less than if we recomputed clockNow.
	  if (delay > 24) {
	    if (time < Infinity) timeout$1 = setTimeout(wake, time - clock.now() - clockSkew);
	    if (interval) interval = clearInterval(interval);
	  } else {
	    if (!interval) clockLast = clock.now(), interval = setInterval(poke, pokeDelay);
	    frame = 1, setFrame(wake);
	  }
	}

	function timeout(callback, delay, time) {
	  var t = new Timer;
	  delay = delay == null ? 0 : +delay;
	  t.restart(function(elapsed) {
	    t.stop();
	    callback(elapsed + delay);
	  }, delay, time);
	  return t;
	}

	var emptyOn = dispatch("start", "end", "cancel", "interrupt");
	var emptyTween = [];

	var CREATED = 0;
	var SCHEDULED = 1;
	var STARTING = 2;
	var STARTED = 3;
	var RUNNING = 4;
	var ENDING = 5;
	var ENDED = 6;

	function schedule(node, name, id, index, group, timing) {
	  var schedules = node.__transition;
	  if (!schedules) node.__transition = {};
	  else if (id in schedules) return;
	  create(node, id, {
	    name: name,
	    index: index, // For context during callback.
	    group: group, // For context during callback.
	    on: emptyOn,
	    tween: emptyTween,
	    time: timing.time,
	    delay: timing.delay,
	    duration: timing.duration,
	    ease: timing.ease,
	    timer: null,
	    state: CREATED
	  });
	}

	function init$1(node, id) {
	  var schedule = get(node, id);
	  if (schedule.state > CREATED) throw new Error("too late; already scheduled");
	  return schedule;
	}

	function set(node, id) {
	  var schedule = get(node, id);
	  if (schedule.state > STARTED) throw new Error("too late; already running");
	  return schedule;
	}

	function get(node, id) {
	  var schedule = node.__transition;
	  if (!schedule || !(schedule = schedule[id])) throw new Error("transition not found");
	  return schedule;
	}

	function create(node, id, self) {
	  var schedules = node.__transition,
	      tween;

	  // Initialize the self timer when the transition is created.
	  // Note the actual delay is not known until the first callback!
	  schedules[id] = self;
	  self.timer = timer(schedule, 0, self.time);

	  function schedule(elapsed) {
	    self.state = SCHEDULED;
	    self.timer.restart(start, self.delay, self.time);

	    // If the elapsed delay is less than our first sleep, start immediately.
	    if (self.delay <= elapsed) start(elapsed - self.delay);
	  }

	  function start(elapsed) {
	    var i, j, n, o;

	    // If the state is not SCHEDULED, then we previously errored on start.
	    if (self.state !== SCHEDULED) return stop();

	    for (i in schedules) {
	      o = schedules[i];
	      if (o.name !== self.name) continue;

	      // While this element already has a starting transition during this frame,
	      // defer starting an interrupting transition until that transition has a
	      // chance to tick (and possibly end); see d3/d3-transition#54!
	      if (o.state === STARTED) return timeout(start);

	      // Interrupt the active transition, if any.
	      if (o.state === RUNNING) {
	        o.state = ENDED;
	        o.timer.stop();
	        o.on.call("interrupt", node, node.__data__, o.index, o.group);
	        delete schedules[i];
	      }

	      // Cancel any pre-empted transitions.
	      else if (+i < id) {
	        o.state = ENDED;
	        o.timer.stop();
	        o.on.call("cancel", node, node.__data__, o.index, o.group);
	        delete schedules[i];
	      }
	    }

	    // Defer the first tick to end of the current frame; see d3/d3#1576.
	    // Note the transition may be canceled after start and before the first tick!
	    // Note this must be scheduled before the start event; see d3/d3-transition#16!
	    // Assuming this is successful, subsequent callbacks go straight to tick.
	    timeout(function() {
	      if (self.state === STARTED) {
	        self.state = RUNNING;
	        self.timer.restart(tick, self.delay, self.time);
	        tick(elapsed);
	      }
	    });

	    // Dispatch the start event.
	    // Note this must be done before the tween are initialized.
	    self.state = STARTING;
	    self.on.call("start", node, node.__data__, self.index, self.group);
	    if (self.state !== STARTING) return; // interrupted
	    self.state = STARTED;

	    // Initialize the tween, deleting null tween.
	    tween = new Array(n = self.tween.length);
	    for (i = 0, j = -1; i < n; ++i) {
	      if (o = self.tween[i].value.call(node, node.__data__, self.index, self.group)) {
	        tween[++j] = o;
	      }
	    }
	    tween.length = j + 1;
	  }

	  function tick(elapsed) {
	    var t = elapsed < self.duration ? self.ease.call(null, elapsed / self.duration) : (self.timer.restart(stop), self.state = ENDING, 1),
	        i = -1,
	        n = tween.length;

	    while (++i < n) {
	      tween[i].call(node, t);
	    }

	    // Dispatch the end event.
	    if (self.state === ENDING) {
	      self.on.call("end", node, node.__data__, self.index, self.group);
	      stop();
	    }
	  }

	  function stop() {
	    self.state = ENDED;
	    self.timer.stop();
	    delete schedules[id];
	    for (var i in schedules) return; // eslint-disable-line no-unused-vars
	    delete node.__transition;
	  }
	}

	function interrupt(node, name) {
	  var schedules = node.__transition,
	      schedule,
	      active,
	      empty = true,
	      i;

	  if (!schedules) return;

	  name = name == null ? null : name + "";

	  for (i in schedules) {
	    if ((schedule = schedules[i]).name !== name) { empty = false; continue; }
	    active = schedule.state > STARTING && schedule.state < ENDING;
	    schedule.state = ENDED;
	    schedule.timer.stop();
	    schedule.on.call(active ? "interrupt" : "cancel", node, node.__data__, schedule.index, schedule.group);
	    delete schedules[i];
	  }

	  if (empty) delete node.__transition;
	}

	function selection_interrupt(name) {
	  return this.each(function() {
	    interrupt(this, name);
	  });
	}

	function define(constructor, factory, prototype) {
	  constructor.prototype = factory.prototype = prototype;
	  prototype.constructor = constructor;
	}

	function extend(parent, definition) {
	  var prototype = Object.create(parent.prototype);
	  for (var key in definition) prototype[key] = definition[key];
	  return prototype;
	}

	function Color() {}

	var darker = 0.7;
	var brighter = 1 / darker;

	var reI = "\\s*([+-]?\\d+)\\s*",
	    reN = "\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)\\s*",
	    reP = "\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)%\\s*",
	    reHex = /^#([0-9a-f]{3,8})$/,
	    reRgbInteger = new RegExp("^rgb\\(" + [reI, reI, reI] + "\\)$"),
	    reRgbPercent = new RegExp("^rgb\\(" + [reP, reP, reP] + "\\)$"),
	    reRgbaInteger = new RegExp("^rgba\\(" + [reI, reI, reI, reN] + "\\)$"),
	    reRgbaPercent = new RegExp("^rgba\\(" + [reP, reP, reP, reN] + "\\)$"),
	    reHslPercent = new RegExp("^hsl\\(" + [reN, reP, reP] + "\\)$"),
	    reHslaPercent = new RegExp("^hsla\\(" + [reN, reP, reP, reN] + "\\)$");

	var named = {
	  aliceblue: 0xf0f8ff,
	  antiquewhite: 0xfaebd7,
	  aqua: 0x00ffff,
	  aquamarine: 0x7fffd4,
	  azure: 0xf0ffff,
	  beige: 0xf5f5dc,
	  bisque: 0xffe4c4,
	  black: 0x000000,
	  blanchedalmond: 0xffebcd,
	  blue: 0x0000ff,
	  blueviolet: 0x8a2be2,
	  brown: 0xa52a2a,
	  burlywood: 0xdeb887,
	  cadetblue: 0x5f9ea0,
	  chartreuse: 0x7fff00,
	  chocolate: 0xd2691e,
	  coral: 0xff7f50,
	  cornflowerblue: 0x6495ed,
	  cornsilk: 0xfff8dc,
	  crimson: 0xdc143c,
	  cyan: 0x00ffff,
	  darkblue: 0x00008b,
	  darkcyan: 0x008b8b,
	  darkgoldenrod: 0xb8860b,
	  darkgray: 0xa9a9a9,
	  darkgreen: 0x006400,
	  darkgrey: 0xa9a9a9,
	  darkkhaki: 0xbdb76b,
	  darkmagenta: 0x8b008b,
	  darkolivegreen: 0x556b2f,
	  darkorange: 0xff8c00,
	  darkorchid: 0x9932cc,
	  darkred: 0x8b0000,
	  darksalmon: 0xe9967a,
	  darkseagreen: 0x8fbc8f,
	  darkslateblue: 0x483d8b,
	  darkslategray: 0x2f4f4f,
	  darkslategrey: 0x2f4f4f,
	  darkturquoise: 0x00ced1,
	  darkviolet: 0x9400d3,
	  deeppink: 0xff1493,
	  deepskyblue: 0x00bfff,
	  dimgray: 0x696969,
	  dimgrey: 0x696969,
	  dodgerblue: 0x1e90ff,
	  firebrick: 0xb22222,
	  floralwhite: 0xfffaf0,
	  forestgreen: 0x228b22,
	  fuchsia: 0xff00ff,
	  gainsboro: 0xdcdcdc,
	  ghostwhite: 0xf8f8ff,
	  gold: 0xffd700,
	  goldenrod: 0xdaa520,
	  gray: 0x808080,
	  green: 0x008000,
	  greenyellow: 0xadff2f,
	  grey: 0x808080,
	  honeydew: 0xf0fff0,
	  hotpink: 0xff69b4,
	  indianred: 0xcd5c5c,
	  indigo: 0x4b0082,
	  ivory: 0xfffff0,
	  khaki: 0xf0e68c,
	  lavender: 0xe6e6fa,
	  lavenderblush: 0xfff0f5,
	  lawngreen: 0x7cfc00,
	  lemonchiffon: 0xfffacd,
	  lightblue: 0xadd8e6,
	  lightcoral: 0xf08080,
	  lightcyan: 0xe0ffff,
	  lightgoldenrodyellow: 0xfafad2,
	  lightgray: 0xd3d3d3,
	  lightgreen: 0x90ee90,
	  lightgrey: 0xd3d3d3,
	  lightpink: 0xffb6c1,
	  lightsalmon: 0xffa07a,
	  lightseagreen: 0x20b2aa,
	  lightskyblue: 0x87cefa,
	  lightslategray: 0x778899,
	  lightslategrey: 0x778899,
	  lightsteelblue: 0xb0c4de,
	  lightyellow: 0xffffe0,
	  lime: 0x00ff00,
	  limegreen: 0x32cd32,
	  linen: 0xfaf0e6,
	  magenta: 0xff00ff,
	  maroon: 0x800000,
	  mediumaquamarine: 0x66cdaa,
	  mediumblue: 0x0000cd,
	  mediumorchid: 0xba55d3,
	  mediumpurple: 0x9370db,
	  mediumseagreen: 0x3cb371,
	  mediumslateblue: 0x7b68ee,
	  mediumspringgreen: 0x00fa9a,
	  mediumturquoise: 0x48d1cc,
	  mediumvioletred: 0xc71585,
	  midnightblue: 0x191970,
	  mintcream: 0xf5fffa,
	  mistyrose: 0xffe4e1,
	  moccasin: 0xffe4b5,
	  navajowhite: 0xffdead,
	  navy: 0x000080,
	  oldlace: 0xfdf5e6,
	  olive: 0x808000,
	  olivedrab: 0x6b8e23,
	  orange: 0xffa500,
	  orangered: 0xff4500,
	  orchid: 0xda70d6,
	  palegoldenrod: 0xeee8aa,
	  palegreen: 0x98fb98,
	  paleturquoise: 0xafeeee,
	  palevioletred: 0xdb7093,
	  papayawhip: 0xffefd5,
	  peachpuff: 0xffdab9,
	  peru: 0xcd853f,
	  pink: 0xffc0cb,
	  plum: 0xdda0dd,
	  powderblue: 0xb0e0e6,
	  purple: 0x800080,
	  rebeccapurple: 0x663399,
	  red: 0xff0000,
	  rosybrown: 0xbc8f8f,
	  royalblue: 0x4169e1,
	  saddlebrown: 0x8b4513,
	  salmon: 0xfa8072,
	  sandybrown: 0xf4a460,
	  seagreen: 0x2e8b57,
	  seashell: 0xfff5ee,
	  sienna: 0xa0522d,
	  silver: 0xc0c0c0,
	  skyblue: 0x87ceeb,
	  slateblue: 0x6a5acd,
	  slategray: 0x708090,
	  slategrey: 0x708090,
	  snow: 0xfffafa,
	  springgreen: 0x00ff7f,
	  steelblue: 0x4682b4,
	  tan: 0xd2b48c,
	  teal: 0x008080,
	  thistle: 0xd8bfd8,
	  tomato: 0xff6347,
	  turquoise: 0x40e0d0,
	  violet: 0xee82ee,
	  wheat: 0xf5deb3,
	  white: 0xffffff,
	  whitesmoke: 0xf5f5f5,
	  yellow: 0xffff00,
	  yellowgreen: 0x9acd32
	};

	define(Color, color, {
	  copy: function(channels) {
	    return Object.assign(new this.constructor, this, channels);
	  },
	  displayable: function() {
	    return this.rgb().displayable();
	  },
	  hex: color_formatHex, // Deprecated! Use color.formatHex.
	  formatHex: color_formatHex,
	  formatHsl: color_formatHsl,
	  formatRgb: color_formatRgb,
	  toString: color_formatRgb
	});

	function color_formatHex() {
	  return this.rgb().formatHex();
	}

	function color_formatHsl() {
	  return hslConvert(this).formatHsl();
	}

	function color_formatRgb() {
	  return this.rgb().formatRgb();
	}

	function color(format) {
	  var m, l;
	  format = (format + "").trim().toLowerCase();
	  return (m = reHex.exec(format)) ? (l = m[1].length, m = parseInt(m[1], 16), l === 6 ? rgbn(m) // #ff0000
	      : l === 3 ? new Rgb((m >> 8 & 0xf) | (m >> 4 & 0xf0), (m >> 4 & 0xf) | (m & 0xf0), ((m & 0xf) << 4) | (m & 0xf), 1) // #f00
	      : l === 8 ? rgba(m >> 24 & 0xff, m >> 16 & 0xff, m >> 8 & 0xff, (m & 0xff) / 0xff) // #ff000000
	      : l === 4 ? rgba((m >> 12 & 0xf) | (m >> 8 & 0xf0), (m >> 8 & 0xf) | (m >> 4 & 0xf0), (m >> 4 & 0xf) | (m & 0xf0), (((m & 0xf) << 4) | (m & 0xf)) / 0xff) // #f000
	      : null) // invalid hex
	      : (m = reRgbInteger.exec(format)) ? new Rgb(m[1], m[2], m[3], 1) // rgb(255, 0, 0)
	      : (m = reRgbPercent.exec(format)) ? new Rgb(m[1] * 255 / 100, m[2] * 255 / 100, m[3] * 255 / 100, 1) // rgb(100%, 0%, 0%)
	      : (m = reRgbaInteger.exec(format)) ? rgba(m[1], m[2], m[3], m[4]) // rgba(255, 0, 0, 1)
	      : (m = reRgbaPercent.exec(format)) ? rgba(m[1] * 255 / 100, m[2] * 255 / 100, m[3] * 255 / 100, m[4]) // rgb(100%, 0%, 0%, 1)
	      : (m = reHslPercent.exec(format)) ? hsla(m[1], m[2] / 100, m[3] / 100, 1) // hsl(120, 50%, 50%)
	      : (m = reHslaPercent.exec(format)) ? hsla(m[1], m[2] / 100, m[3] / 100, m[4]) // hsla(120, 50%, 50%, 1)
	      : named.hasOwnProperty(format) ? rgbn(named[format]) // eslint-disable-line no-prototype-builtins
	      : format === "transparent" ? new Rgb(NaN, NaN, NaN, 0)
	      : null;
	}

	function rgbn(n) {
	  return new Rgb(n >> 16 & 0xff, n >> 8 & 0xff, n & 0xff, 1);
	}

	function rgba(r, g, b, a) {
	  if (a <= 0) r = g = b = NaN;
	  return new Rgb(r, g, b, a);
	}

	function rgbConvert(o) {
	  if (!(o instanceof Color)) o = color(o);
	  if (!o) return new Rgb;
	  o = o.rgb();
	  return new Rgb(o.r, o.g, o.b, o.opacity);
	}

	function rgb(r, g, b, opacity) {
	  return arguments.length === 1 ? rgbConvert(r) : new Rgb(r, g, b, opacity == null ? 1 : opacity);
	}

	function Rgb(r, g, b, opacity) {
	  this.r = +r;
	  this.g = +g;
	  this.b = +b;
	  this.opacity = +opacity;
	}

	define(Rgb, rgb, extend(Color, {
	  brighter: function(k) {
	    k = k == null ? brighter : Math.pow(brighter, k);
	    return new Rgb(this.r * k, this.g * k, this.b * k, this.opacity);
	  },
	  darker: function(k) {
	    k = k == null ? darker : Math.pow(darker, k);
	    return new Rgb(this.r * k, this.g * k, this.b * k, this.opacity);
	  },
	  rgb: function() {
	    return this;
	  },
	  displayable: function() {
	    return (-0.5 <= this.r && this.r < 255.5)
	        && (-0.5 <= this.g && this.g < 255.5)
	        && (-0.5 <= this.b && this.b < 255.5)
	        && (0 <= this.opacity && this.opacity <= 1);
	  },
	  hex: rgb_formatHex, // Deprecated! Use color.formatHex.
	  formatHex: rgb_formatHex,
	  formatRgb: rgb_formatRgb,
	  toString: rgb_formatRgb
	}));

	function rgb_formatHex() {
	  return "#" + hex(this.r) + hex(this.g) + hex(this.b);
	}

	function rgb_formatRgb() {
	  var a = this.opacity; a = isNaN(a) ? 1 : Math.max(0, Math.min(1, a));
	  return (a === 1 ? "rgb(" : "rgba(")
	      + Math.max(0, Math.min(255, Math.round(this.r) || 0)) + ", "
	      + Math.max(0, Math.min(255, Math.round(this.g) || 0)) + ", "
	      + Math.max(0, Math.min(255, Math.round(this.b) || 0))
	      + (a === 1 ? ")" : ", " + a + ")");
	}

	function hex(value) {
	  value = Math.max(0, Math.min(255, Math.round(value) || 0));
	  return (value < 16 ? "0" : "") + value.toString(16);
	}

	function hsla(h, s, l, a) {
	  if (a <= 0) h = s = l = NaN;
	  else if (l <= 0 || l >= 1) h = s = NaN;
	  else if (s <= 0) h = NaN;
	  return new Hsl(h, s, l, a);
	}

	function hslConvert(o) {
	  if (o instanceof Hsl) return new Hsl(o.h, o.s, o.l, o.opacity);
	  if (!(o instanceof Color)) o = color(o);
	  if (!o) return new Hsl;
	  if (o instanceof Hsl) return o;
	  o = o.rgb();
	  var r = o.r / 255,
	      g = o.g / 255,
	      b = o.b / 255,
	      min = Math.min(r, g, b),
	      max = Math.max(r, g, b),
	      h = NaN,
	      s = max - min,
	      l = (max + min) / 2;
	  if (s) {
	    if (r === max) h = (g - b) / s + (g < b) * 6;
	    else if (g === max) h = (b - r) / s + 2;
	    else h = (r - g) / s + 4;
	    s /= l < 0.5 ? max + min : 2 - max - min;
	    h *= 60;
	  } else {
	    s = l > 0 && l < 1 ? 0 : h;
	  }
	  return new Hsl(h, s, l, o.opacity);
	}

	function hsl(h, s, l, opacity) {
	  return arguments.length === 1 ? hslConvert(h) : new Hsl(h, s, l, opacity == null ? 1 : opacity);
	}

	function Hsl(h, s, l, opacity) {
	  this.h = +h;
	  this.s = +s;
	  this.l = +l;
	  this.opacity = +opacity;
	}

	define(Hsl, hsl, extend(Color, {
	  brighter: function(k) {
	    k = k == null ? brighter : Math.pow(brighter, k);
	    return new Hsl(this.h, this.s, this.l * k, this.opacity);
	  },
	  darker: function(k) {
	    k = k == null ? darker : Math.pow(darker, k);
	    return new Hsl(this.h, this.s, this.l * k, this.opacity);
	  },
	  rgb: function() {
	    var h = this.h % 360 + (this.h < 0) * 360,
	        s = isNaN(h) || isNaN(this.s) ? 0 : this.s,
	        l = this.l,
	        m2 = l + (l < 0.5 ? l : 1 - l) * s,
	        m1 = 2 * l - m2;
	    return new Rgb(
	      hsl2rgb(h >= 240 ? h - 240 : h + 120, m1, m2),
	      hsl2rgb(h, m1, m2),
	      hsl2rgb(h < 120 ? h + 240 : h - 120, m1, m2),
	      this.opacity
	    );
	  },
	  displayable: function() {
	    return (0 <= this.s && this.s <= 1 || isNaN(this.s))
	        && (0 <= this.l && this.l <= 1)
	        && (0 <= this.opacity && this.opacity <= 1);
	  },
	  formatHsl: function() {
	    var a = this.opacity; a = isNaN(a) ? 1 : Math.max(0, Math.min(1, a));
	    return (a === 1 ? "hsl(" : "hsla(")
	        + (this.h || 0) + ", "
	        + (this.s || 0) * 100 + "%, "
	        + (this.l || 0) * 100 + "%"
	        + (a === 1 ? ")" : ", " + a + ")");
	  }
	}));

	/* From FvD 13.37, CSS Color Module Level 3 */
	function hsl2rgb(h, m1, m2) {
	  return (h < 60 ? m1 + (m2 - m1) * h / 60
	      : h < 180 ? m2
	      : h < 240 ? m1 + (m2 - m1) * (240 - h) / 60
	      : m1) * 255;
	}

	function constant(x) {
	  return function() {
	    return x;
	  };
	}

	function linear(a, d) {
	  return function(t) {
	    return a + t * d;
	  };
	}

	function exponential(a, b, y) {
	  return a = Math.pow(a, y), b = Math.pow(b, y) - a, y = 1 / y, function(t) {
	    return Math.pow(a + t * b, y);
	  };
	}

	function gamma(y) {
	  return (y = +y) === 1 ? nogamma : function(a, b) {
	    return b - a ? exponential(a, b, y) : constant(isNaN(a) ? b : a);
	  };
	}

	function nogamma(a, b) {
	  var d = b - a;
	  return d ? linear(a, d) : constant(isNaN(a) ? b : a);
	}

	var interpolateRgb = (function rgbGamma(y) {
	  var color = gamma(y);

	  function rgb$1(start, end) {
	    var r = color((start = rgb(start)).r, (end = rgb(end)).r),
	        g = color(start.g, end.g),
	        b = color(start.b, end.b),
	        opacity = nogamma(start.opacity, end.opacity);
	    return function(t) {
	      start.r = r(t);
	      start.g = g(t);
	      start.b = b(t);
	      start.opacity = opacity(t);
	      return start + "";
	    };
	  }

	  rgb$1.gamma = rgbGamma;

	  return rgb$1;
	})(1);

	function interpolateNumber(a, b) {
	  return a = +a, b = +b, function(t) {
	    return a * (1 - t) + b * t;
	  };
	}

	var reA = /[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,
	    reB = new RegExp(reA.source, "g");

	function zero(b) {
	  return function() {
	    return b;
	  };
	}

	function one(b) {
	  return function(t) {
	    return b(t) + "";
	  };
	}

	function interpolateString(a, b) {
	  var bi = reA.lastIndex = reB.lastIndex = 0, // scan index for next number in b
	      am, // current match in a
	      bm, // current match in b
	      bs, // string preceding current number in b, if any
	      i = -1, // index in s
	      s = [], // string constants and placeholders
	      q = []; // number interpolators

	  // Coerce inputs to strings.
	  a = a + "", b = b + "";

	  // Interpolate pairs of numbers in a & b.
	  while ((am = reA.exec(a))
	      && (bm = reB.exec(b))) {
	    if ((bs = bm.index) > bi) { // a string precedes the next number in b
	      bs = b.slice(bi, bs);
	      if (s[i]) s[i] += bs; // coalesce with previous string
	      else s[++i] = bs;
	    }
	    if ((am = am[0]) === (bm = bm[0])) { // numbers in a & b match
	      if (s[i]) s[i] += bm; // coalesce with previous string
	      else s[++i] = bm;
	    } else { // interpolate non-matching numbers
	      s[++i] = null;
	      q.push({i: i, x: interpolateNumber(am, bm)});
	    }
	    bi = reB.lastIndex;
	  }

	  // Add remains of b.
	  if (bi < b.length) {
	    bs = b.slice(bi);
	    if (s[i]) s[i] += bs; // coalesce with previous string
	    else s[++i] = bs;
	  }

	  // Special optimization for only a single match.
	  // Otherwise, interpolate each of the numbers and rejoin the string.
	  return s.length < 2 ? (q[0]
	      ? one(q[0].x)
	      : zero(b))
	      : (b = q.length, function(t) {
	          for (var i = 0, o; i < b; ++i) s[(o = q[i]).i] = o.x(t);
	          return s.join("");
	        });
	}

	var degrees = 180 / Math.PI;

	var identity = {
	  translateX: 0,
	  translateY: 0,
	  rotate: 0,
	  skewX: 0,
	  scaleX: 1,
	  scaleY: 1
	};

	function decompose(a, b, c, d, e, f) {
	  var scaleX, scaleY, skewX;
	  if (scaleX = Math.sqrt(a * a + b * b)) a /= scaleX, b /= scaleX;
	  if (skewX = a * c + b * d) c -= a * skewX, d -= b * skewX;
	  if (scaleY = Math.sqrt(c * c + d * d)) c /= scaleY, d /= scaleY, skewX /= scaleY;
	  if (a * d < b * c) a = -a, b = -b, skewX = -skewX, scaleX = -scaleX;
	  return {
	    translateX: e,
	    translateY: f,
	    rotate: Math.atan2(b, a) * degrees,
	    skewX: Math.atan(skewX) * degrees,
	    scaleX: scaleX,
	    scaleY: scaleY
	  };
	}

	var cssNode,
	    cssRoot,
	    cssView,
	    svgNode;

	function parseCss(value) {
	  if (value === "none") return identity;
	  if (!cssNode) cssNode = document.createElement("DIV"), cssRoot = document.documentElement, cssView = document.defaultView;
	  cssNode.style.transform = value;
	  value = cssView.getComputedStyle(cssRoot.appendChild(cssNode), null).getPropertyValue("transform");
	  cssRoot.removeChild(cssNode);
	  value = value.slice(7, -1).split(",");
	  return decompose(+value[0], +value[1], +value[2], +value[3], +value[4], +value[5]);
	}

	function parseSvg(value) {
	  if (value == null) return identity;
	  if (!svgNode) svgNode = document.createElementNS("http://www.w3.org/2000/svg", "g");
	  svgNode.setAttribute("transform", value);
	  if (!(value = svgNode.transform.baseVal.consolidate())) return identity;
	  value = value.matrix;
	  return decompose(value.a, value.b, value.c, value.d, value.e, value.f);
	}

	function interpolateTransform(parse, pxComma, pxParen, degParen) {

	  function pop(s) {
	    return s.length ? s.pop() + " " : "";
	  }

	  function translate(xa, ya, xb, yb, s, q) {
	    if (xa !== xb || ya !== yb) {
	      var i = s.push("translate(", null, pxComma, null, pxParen);
	      q.push({i: i - 4, x: interpolateNumber(xa, xb)}, {i: i - 2, x: interpolateNumber(ya, yb)});
	    } else if (xb || yb) {
	      s.push("translate(" + xb + pxComma + yb + pxParen);
	    }
	  }

	  function rotate(a, b, s, q) {
	    if (a !== b) {
	      if (a - b > 180) b += 360; else if (b - a > 180) a += 360; // shortest path
	      q.push({i: s.push(pop(s) + "rotate(", null, degParen) - 2, x: interpolateNumber(a, b)});
	    } else if (b) {
	      s.push(pop(s) + "rotate(" + b + degParen);
	    }
	  }

	  function skewX(a, b, s, q) {
	    if (a !== b) {
	      q.push({i: s.push(pop(s) + "skewX(", null, degParen) - 2, x: interpolateNumber(a, b)});
	    } else if (b) {
	      s.push(pop(s) + "skewX(" + b + degParen);
	    }
	  }

	  function scale(xa, ya, xb, yb, s, q) {
	    if (xa !== xb || ya !== yb) {
	      var i = s.push(pop(s) + "scale(", null, ",", null, ")");
	      q.push({i: i - 4, x: interpolateNumber(xa, xb)}, {i: i - 2, x: interpolateNumber(ya, yb)});
	    } else if (xb !== 1 || yb !== 1) {
	      s.push(pop(s) + "scale(" + xb + "," + yb + ")");
	    }
	  }

	  return function(a, b) {
	    var s = [], // string constants and placeholders
	        q = []; // number interpolators
	    a = parse(a), b = parse(b);
	    translate(a.translateX, a.translateY, b.translateX, b.translateY, s, q);
	    rotate(a.rotate, b.rotate, s, q);
	    skewX(a.skewX, b.skewX, s, q);
	    scale(a.scaleX, a.scaleY, b.scaleX, b.scaleY, s, q);
	    a = b = null; // gc
	    return function(t) {
	      var i = -1, n = q.length, o;
	      while (++i < n) s[(o = q[i]).i] = o.x(t);
	      return s.join("");
	    };
	  };
	}

	var interpolateTransformCss = interpolateTransform(parseCss, "px, ", "px)", "deg)");
	var interpolateTransformSvg = interpolateTransform(parseSvg, ", ", ")", ")");

	function tweenRemove(id, name) {
	  var tween0, tween1;
	  return function() {
	    var schedule = set(this, id),
	        tween = schedule.tween;

	    // If this node shared tween with the previous node,
	    // just assign the updated shared tween and we’re done!
	    // Otherwise, copy-on-write.
	    if (tween !== tween0) {
	      tween1 = tween0 = tween;
	      for (var i = 0, n = tween1.length; i < n; ++i) {
	        if (tween1[i].name === name) {
	          tween1 = tween1.slice();
	          tween1.splice(i, 1);
	          break;
	        }
	      }
	    }

	    schedule.tween = tween1;
	  };
	}

	function tweenFunction(id, name, value) {
	  var tween0, tween1;
	  if (typeof value !== "function") throw new Error;
	  return function() {
	    var schedule = set(this, id),
	        tween = schedule.tween;

	    // If this node shared tween with the previous node,
	    // just assign the updated shared tween and we’re done!
	    // Otherwise, copy-on-write.
	    if (tween !== tween0) {
	      tween1 = (tween0 = tween).slice();
	      for (var t = {name: name, value: value}, i = 0, n = tween1.length; i < n; ++i) {
	        if (tween1[i].name === name) {
	          tween1[i] = t;
	          break;
	        }
	      }
	      if (i === n) tween1.push(t);
	    }

	    schedule.tween = tween1;
	  };
	}

	function transition_tween(name, value) {
	  var id = this._id;

	  name += "";

	  if (arguments.length < 2) {
	    var tween = get(this.node(), id).tween;
	    for (var i = 0, n = tween.length, t; i < n; ++i) {
	      if ((t = tween[i]).name === name) {
	        return t.value;
	      }
	    }
	    return null;
	  }

	  return this.each((value == null ? tweenRemove : tweenFunction)(id, name, value));
	}

	function tweenValue(transition, name, value) {
	  var id = transition._id;

	  transition.each(function() {
	    var schedule = set(this, id);
	    (schedule.value || (schedule.value = {}))[name] = value.apply(this, arguments);
	  });

	  return function(node) {
	    return get(node, id).value[name];
	  };
	}

	function interpolate(a, b) {
	  var c;
	  return (typeof b === "number" ? interpolateNumber
	      : b instanceof color ? interpolateRgb
	      : (c = color(b)) ? (b = c, interpolateRgb)
	      : interpolateString)(a, b);
	}

	function attrRemove(name) {
	  return function() {
	    this.removeAttribute(name);
	  };
	}

	function attrRemoveNS(fullname) {
	  return function() {
	    this.removeAttributeNS(fullname.space, fullname.local);
	  };
	}

	function attrConstant(name, interpolate, value1) {
	  var string00,
	      string1 = value1 + "",
	      interpolate0;
	  return function() {
	    var string0 = this.getAttribute(name);
	    return string0 === string1 ? null
	        : string0 === string00 ? interpolate0
	        : interpolate0 = interpolate(string00 = string0, value1);
	  };
	}

	function attrConstantNS(fullname, interpolate, value1) {
	  var string00,
	      string1 = value1 + "",
	      interpolate0;
	  return function() {
	    var string0 = this.getAttributeNS(fullname.space, fullname.local);
	    return string0 === string1 ? null
	        : string0 === string00 ? interpolate0
	        : interpolate0 = interpolate(string00 = string0, value1);
	  };
	}

	function attrFunction(name, interpolate, value) {
	  var string00,
	      string10,
	      interpolate0;
	  return function() {
	    var string0, value1 = value(this), string1;
	    if (value1 == null) return void this.removeAttribute(name);
	    string0 = this.getAttribute(name);
	    string1 = value1 + "";
	    return string0 === string1 ? null
	        : string0 === string00 && string1 === string10 ? interpolate0
	        : (string10 = string1, interpolate0 = interpolate(string00 = string0, value1));
	  };
	}

	function attrFunctionNS(fullname, interpolate, value) {
	  var string00,
	      string10,
	      interpolate0;
	  return function() {
	    var string0, value1 = value(this), string1;
	    if (value1 == null) return void this.removeAttributeNS(fullname.space, fullname.local);
	    string0 = this.getAttributeNS(fullname.space, fullname.local);
	    string1 = value1 + "";
	    return string0 === string1 ? null
	        : string0 === string00 && string1 === string10 ? interpolate0
	        : (string10 = string1, interpolate0 = interpolate(string00 = string0, value1));
	  };
	}

	function transition_attr(name, value) {
	  var fullname = namespace(name), i = fullname === "transform" ? interpolateTransformSvg : interpolate;
	  return this.attrTween(name, typeof value === "function"
	      ? (fullname.local ? attrFunctionNS : attrFunction)(fullname, i, tweenValue(this, "attr." + name, value))
	      : value == null ? (fullname.local ? attrRemoveNS : attrRemove)(fullname)
	      : (fullname.local ? attrConstantNS : attrConstant)(fullname, i, value));
	}

	function attrInterpolate(name, i) {
	  return function(t) {
	    this.setAttribute(name, i.call(this, t));
	  };
	}

	function attrInterpolateNS(fullname, i) {
	  return function(t) {
	    this.setAttributeNS(fullname.space, fullname.local, i.call(this, t));
	  };
	}

	function attrTweenNS(fullname, value) {
	  var t0, i0;
	  function tween() {
	    var i = value.apply(this, arguments);
	    if (i !== i0) t0 = (i0 = i) && attrInterpolateNS(fullname, i);
	    return t0;
	  }
	  tween._value = value;
	  return tween;
	}

	function attrTween(name, value) {
	  var t0, i0;
	  function tween() {
	    var i = value.apply(this, arguments);
	    if (i !== i0) t0 = (i0 = i) && attrInterpolate(name, i);
	    return t0;
	  }
	  tween._value = value;
	  return tween;
	}

	function transition_attrTween(name, value) {
	  var key = "attr." + name;
	  if (arguments.length < 2) return (key = this.tween(key)) && key._value;
	  if (value == null) return this.tween(key, null);
	  if (typeof value !== "function") throw new Error;
	  var fullname = namespace(name);
	  return this.tween(key, (fullname.local ? attrTweenNS : attrTween)(fullname, value));
	}

	function delayFunction(id, value) {
	  return function() {
	    init$1(this, id).delay = +value.apply(this, arguments);
	  };
	}

	function delayConstant(id, value) {
	  return value = +value, function() {
	    init$1(this, id).delay = value;
	  };
	}

	function transition_delay(value) {
	  var id = this._id;

	  return arguments.length
	      ? this.each((typeof value === "function"
	          ? delayFunction
	          : delayConstant)(id, value))
	      : get(this.node(), id).delay;
	}

	function durationFunction(id, value) {
	  return function() {
	    set(this, id).duration = +value.apply(this, arguments);
	  };
	}

	function durationConstant(id, value) {
	  return value = +value, function() {
	    set(this, id).duration = value;
	  };
	}

	function transition_duration(value) {
	  var id = this._id;

	  return arguments.length
	      ? this.each((typeof value === "function"
	          ? durationFunction
	          : durationConstant)(id, value))
	      : get(this.node(), id).duration;
	}

	function easeConstant(id, value) {
	  if (typeof value !== "function") throw new Error;
	  return function() {
	    set(this, id).ease = value;
	  };
	}

	function transition_ease(value) {
	  var id = this._id;

	  return arguments.length
	      ? this.each(easeConstant(id, value))
	      : get(this.node(), id).ease;
	}

	function easeVarying(id, value) {
	  return function() {
	    var v = value.apply(this, arguments);
	    if (typeof v !== "function") throw new Error;
	    set(this, id).ease = v;
	  };
	}

	function transition_easeVarying(value) {
	  if (typeof value !== "function") throw new Error;
	  return this.each(easeVarying(this._id, value));
	}

	function transition_filter(match) {
	  if (typeof match !== "function") match = matcher(match);

	  for (var groups = this._groups, m = groups.length, subgroups = new Array(m), j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, subgroup = subgroups[j] = [], node, i = 0; i < n; ++i) {
	      if ((node = group[i]) && match.call(node, node.__data__, i, group)) {
	        subgroup.push(node);
	      }
	    }
	  }

	  return new Transition(subgroups, this._parents, this._name, this._id);
	}

	function transition_merge(transition) {
	  if (transition._id !== this._id) throw new Error;

	  for (var groups0 = this._groups, groups1 = transition._groups, m0 = groups0.length, m1 = groups1.length, m = Math.min(m0, m1), merges = new Array(m0), j = 0; j < m; ++j) {
	    for (var group0 = groups0[j], group1 = groups1[j], n = group0.length, merge = merges[j] = new Array(n), node, i = 0; i < n; ++i) {
	      if (node = group0[i] || group1[i]) {
	        merge[i] = node;
	      }
	    }
	  }

	  for (; j < m0; ++j) {
	    merges[j] = groups0[j];
	  }

	  return new Transition(merges, this._parents, this._name, this._id);
	}

	function start(name) {
	  return (name + "").trim().split(/^|\s+/).every(function(t) {
	    var i = t.indexOf(".");
	    if (i >= 0) t = t.slice(0, i);
	    return !t || t === "start";
	  });
	}

	function onFunction(id, name, listener) {
	  var on0, on1, sit = start(name) ? init$1 : set;
	  return function() {
	    var schedule = sit(this, id),
	        on = schedule.on;

	    // If this node shared a dispatch with the previous node,
	    // just assign the updated shared dispatch and we’re done!
	    // Otherwise, copy-on-write.
	    if (on !== on0) (on1 = (on0 = on).copy()).on(name, listener);

	    schedule.on = on1;
	  };
	}

	function transition_on(name, listener) {
	  var id = this._id;

	  return arguments.length < 2
	      ? get(this.node(), id).on.on(name)
	      : this.each(onFunction(id, name, listener));
	}

	function removeFunction(id) {
	  return function() {
	    var parent = this.parentNode;
	    for (var i in this.__transition) if (+i !== id) return;
	    if (parent) parent.removeChild(this);
	  };
	}

	function transition_remove() {
	  return this.on("end.remove", removeFunction(this._id));
	}

	function transition_select(select) {
	  var name = this._name,
	      id = this._id;

	  if (typeof select !== "function") select = selector(select);

	  for (var groups = this._groups, m = groups.length, subgroups = new Array(m), j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, subgroup = subgroups[j] = new Array(n), node, subnode, i = 0; i < n; ++i) {
	      if ((node = group[i]) && (subnode = select.call(node, node.__data__, i, group))) {
	        if ("__data__" in node) subnode.__data__ = node.__data__;
	        subgroup[i] = subnode;
	        schedule(subgroup[i], name, id, i, subgroup, get(node, id));
	      }
	    }
	  }

	  return new Transition(subgroups, this._parents, name, id);
	}

	function transition_selectAll(select) {
	  var name = this._name,
	      id = this._id;

	  if (typeof select !== "function") select = selectorAll(select);

	  for (var groups = this._groups, m = groups.length, subgroups = [], parents = [], j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, node, i = 0; i < n; ++i) {
	      if (node = group[i]) {
	        for (var children = select.call(node, node.__data__, i, group), child, inherit = get(node, id), k = 0, l = children.length; k < l; ++k) {
	          if (child = children[k]) {
	            schedule(child, name, id, k, children, inherit);
	          }
	        }
	        subgroups.push(children);
	        parents.push(node);
	      }
	    }
	  }

	  return new Transition(subgroups, parents, name, id);
	}

	var Selection = selection.prototype.constructor;

	function transition_selection() {
	  return new Selection(this._groups, this._parents);
	}

	function styleNull(name, interpolate) {
	  var string00,
	      string10,
	      interpolate0;
	  return function() {
	    var string0 = styleValue(this, name),
	        string1 = (this.style.removeProperty(name), styleValue(this, name));
	    return string0 === string1 ? null
	        : string0 === string00 && string1 === string10 ? interpolate0
	        : interpolate0 = interpolate(string00 = string0, string10 = string1);
	  };
	}

	function styleRemove(name) {
	  return function() {
	    this.style.removeProperty(name);
	  };
	}

	function styleConstant(name, interpolate, value1) {
	  var string00,
	      string1 = value1 + "",
	      interpolate0;
	  return function() {
	    var string0 = styleValue(this, name);
	    return string0 === string1 ? null
	        : string0 === string00 ? interpolate0
	        : interpolate0 = interpolate(string00 = string0, value1);
	  };
	}

	function styleFunction(name, interpolate, value) {
	  var string00,
	      string10,
	      interpolate0;
	  return function() {
	    var string0 = styleValue(this, name),
	        value1 = value(this),
	        string1 = value1 + "";
	    if (value1 == null) string1 = value1 = (this.style.removeProperty(name), styleValue(this, name));
	    return string0 === string1 ? null
	        : string0 === string00 && string1 === string10 ? interpolate0
	        : (string10 = string1, interpolate0 = interpolate(string00 = string0, value1));
	  };
	}

	function styleMaybeRemove(id, name) {
	  var on0, on1, listener0, key = "style." + name, event = "end." + key, remove;
	  return function() {
	    var schedule = set(this, id),
	        on = schedule.on,
	        listener = schedule.value[key] == null ? remove || (remove = styleRemove(name)) : undefined;

	    // If this node shared a dispatch with the previous node,
	    // just assign the updated shared dispatch and we’re done!
	    // Otherwise, copy-on-write.
	    if (on !== on0 || listener0 !== listener) (on1 = (on0 = on).copy()).on(event, listener0 = listener);

	    schedule.on = on1;
	  };
	}

	function transition_style(name, value, priority) {
	  var i = (name += "") === "transform" ? interpolateTransformCss : interpolate;
	  return value == null ? this
	      .styleTween(name, styleNull(name, i))
	      .on("end.style." + name, styleRemove(name))
	    : typeof value === "function" ? this
	      .styleTween(name, styleFunction(name, i, tweenValue(this, "style." + name, value)))
	      .each(styleMaybeRemove(this._id, name))
	    : this
	      .styleTween(name, styleConstant(name, i, value), priority)
	      .on("end.style." + name, null);
	}

	function styleInterpolate(name, i, priority) {
	  return function(t) {
	    this.style.setProperty(name, i.call(this, t), priority);
	  };
	}

	function styleTween(name, value, priority) {
	  var t, i0;
	  function tween() {
	    var i = value.apply(this, arguments);
	    if (i !== i0) t = (i0 = i) && styleInterpolate(name, i, priority);
	    return t;
	  }
	  tween._value = value;
	  return tween;
	}

	function transition_styleTween(name, value, priority) {
	  var key = "style." + (name += "");
	  if (arguments.length < 2) return (key = this.tween(key)) && key._value;
	  if (value == null) return this.tween(key, null);
	  if (typeof value !== "function") throw new Error;
	  return this.tween(key, styleTween(name, value, priority == null ? "" : priority));
	}

	function textConstant(value) {
	  return function() {
	    this.textContent = value;
	  };
	}

	function textFunction(value) {
	  return function() {
	    var value1 = value(this);
	    this.textContent = value1 == null ? "" : value1;
	  };
	}

	function transition_text(value) {
	  return this.tween("text", typeof value === "function"
	      ? textFunction(tweenValue(this, "text", value))
	      : textConstant(value == null ? "" : value + ""));
	}

	function textInterpolate(i) {
	  return function(t) {
	    this.textContent = i.call(this, t);
	  };
	}

	function textTween(value) {
	  var t0, i0;
	  function tween() {
	    var i = value.apply(this, arguments);
	    if (i !== i0) t0 = (i0 = i) && textInterpolate(i);
	    return t0;
	  }
	  tween._value = value;
	  return tween;
	}

	function transition_textTween(value) {
	  var key = "text";
	  if (arguments.length < 1) return (key = this.tween(key)) && key._value;
	  if (value == null) return this.tween(key, null);
	  if (typeof value !== "function") throw new Error;
	  return this.tween(key, textTween(value));
	}

	function transition_transition() {
	  var name = this._name,
	      id0 = this._id,
	      id1 = newId();

	  for (var groups = this._groups, m = groups.length, j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, node, i = 0; i < n; ++i) {
	      if (node = group[i]) {
	        var inherit = get(node, id0);
	        schedule(node, name, id1, i, group, {
	          time: inherit.time + inherit.delay + inherit.duration,
	          delay: 0,
	          duration: inherit.duration,
	          ease: inherit.ease
	        });
	      }
	    }
	  }

	  return new Transition(groups, this._parents, name, id1);
	}

	function transition_end() {
	  var on0, on1, that = this, id = that._id, size = that.size();
	  return new Promise(function(resolve, reject) {
	    var cancel = {value: reject},
	        end = {value: function() { if (--size === 0) resolve(); }};

	    that.each(function() {
	      var schedule = set(this, id),
	          on = schedule.on;

	      // If this node shared a dispatch with the previous node,
	      // just assign the updated shared dispatch and we’re done!
	      // Otherwise, copy-on-write.
	      if (on !== on0) {
	        on1 = (on0 = on).copy();
	        on1._.cancel.push(cancel);
	        on1._.interrupt.push(cancel);
	        on1._.end.push(end);
	      }

	      schedule.on = on1;
	    });

	    // The selection was empty, resolve end immediately
	    if (size === 0) resolve();
	  });
	}

	var id = 0;

	function Transition(groups, parents, name, id) {
	  this._groups = groups;
	  this._parents = parents;
	  this._name = name;
	  this._id = id;
	}

	function newId() {
	  return ++id;
	}

	var selection_prototype = selection.prototype;

	Transition.prototype = {
	  constructor: Transition,
	  select: transition_select,
	  selectAll: transition_selectAll,
	  filter: transition_filter,
	  merge: transition_merge,
	  selection: transition_selection,
	  transition: transition_transition,
	  call: selection_prototype.call,
	  nodes: selection_prototype.nodes,
	  node: selection_prototype.node,
	  size: selection_prototype.size,
	  empty: selection_prototype.empty,
	  each: selection_prototype.each,
	  on: transition_on,
	  attr: transition_attr,
	  attrTween: transition_attrTween,
	  style: transition_style,
	  styleTween: transition_styleTween,
	  text: transition_text,
	  textTween: transition_textTween,
	  remove: transition_remove,
	  tween: transition_tween,
	  delay: transition_delay,
	  duration: transition_duration,
	  ease: transition_ease,
	  easeVarying: transition_easeVarying,
	  end: transition_end,
	  [Symbol.iterator]: selection_prototype[Symbol.iterator]
	};

	function cubicInOut(t) {
	  return ((t *= 2) <= 1 ? t * t * t : (t -= 2) * t * t + 2) / 2;
	}

	var defaultTiming = {
	  time: null, // Set on use.
	  delay: 0,
	  duration: 250,
	  ease: cubicInOut
	};

	function inherit(node, id) {
	  var timing;
	  while (!(timing = node.__transition) || !(timing = timing[id])) {
	    if (!(node = node.parentNode)) {
	      throw new Error(`transition ${id} not found`);
	    }
	  }
	  return timing;
	}

	function selection_transition(name) {
	  var id,
	      timing;

	  if (name instanceof Transition) {
	    id = name._id, name = name._name;
	  } else {
	    id = newId(), (timing = defaultTiming).time = now(), name = name == null ? null : name + "";
	  }

	  for (var groups = this._groups, m = groups.length, j = 0; j < m; ++j) {
	    for (var group = groups[j], n = group.length, node, i = 0; i < n; ++i) {
	      if (node = group[i]) {
	        schedule(node, name, id, i, group, timing || inherit(node, id));
	      }
	    }
	  }

	  return new Transition(groups, this._parents, name, id);
	}

	selection.prototype.interrupt = selection_interrupt;
	selection.prototype.transition = selection_transition;

	function getTypeFromChartLayout (data) {
		switch (true) {
			case (data.string_array):
				return "string";
			case (data.numeric_array):
				return "number";
			case (data.datetime_array):
				return "datetime";
			default:
				return "no type found";
		}
	}

	var window_width, base_size;

	function remToPx(rem) {
		if (window.innerWidth !== window_width) {
			window_width = window.innerWidth;
			base_size = parseFloat(getComputedStyle(document.documentElement).fontSize);
		}
		return rem * base_size;
	}

	function isTextRtl () {
		return document.body.getAttribute("dir") === "rtl";
	}

	function getColorOrTexture (d, texture_type, x_or_y) {
		if (texture_type === "none") return d.area_color;
		else if (d.custom_pattern_i !== null) return `url(#fl-pattern-custom-${d.custom_pattern_i})`;
		else return `url(#${texture_type}-${x_or_y})`;
	}

	function max(values, valueof) {
	  let max;
	  if (valueof === undefined) {
	    for (const value of values) {
	      if (value != null
	          && (max < value || (max === undefined && value >= value))) {
	        max = value;
	      }
	    }
	  } else {
	    let index = -1;
	    for (let value of values) {
	      if ((value = valueof(value, ++index, values)) != null
	          && (max < value || (max === undefined && value >= value))) {
	        max = value;
	      }
	    }
	  }
	  return max;
	}

	function min(values, valueof) {
	  let min;
	  if (valueof === undefined) {
	    for (const value of values) {
	      if (value != null
	          && (min > value || (min === undefined && value >= value))) {
	        min = value;
	      }
	    }
	  } else {
	    let index = -1;
	    for (let value of values) {
	      if ((value = valueof(value, ++index, values)) != null
	          && (min > value || (min === undefined && value >= value))) {
	        min = value;
	      }
	    }
	  }
	  return min;
	}

	((function() {
		var context = document.createElement("canvas").getContext("2d");
		return function(text, font) {
			context.font = font || "10px sans-serif";
			var metrics = context.measureText(text);
			return metrics.width;
		};
	}))();

	function wrapStringToLines(text, font_styles, text_max_lines, max_width) {
		var lines = [];
		var max_line_width = 0;

		// TODO. Should this be just one canvas rather than one per label
		var canvas = document.createElement("canvas");
		var ctx = canvas.getContext("2d");
		ctx.font = font_styles;

		//  if everything fits in one line, so leave as a single string
		var text_width = ctx.measureText(text).width;
		if (text_width <= max_width) {
			lines.push(text);
			max_line_width = text_width;
		}

		// Otherwise…
		else {
			// Create array of words
			var words = text.trim().split(/\s+/g);
			// Loop through the words, adding them until they won't fit
			var current_line = "";
			for (var i = 0; i < words.length; i++) {
				var word = words[i];
				var string = current_line + (current_line ? " " : "") + word;
				var string_width = ctx.measureText(string).width;
				// If the string including the next word fits, update the
				// current_line and the max_width
				if (string_width <= max_width) {
					current_line = string;
					max_line_width = Math.max(max_line_width, string_width);
				}
				// If we reach this point, the text has overflowed,
				// so should always result in a truncation or a new line
				else {
					var last_line = lines.length + 1 == text_max_lines;
					if (!last_line && current_line) {
						lines.push(current_line);
						var word_width = ctx.measureText(word).width;
						if (word_width <= max_width) {
							current_line = word;
							continue;
						}
					}
					// Truncate
					var remove_counter = 1;
					var truncated_string = string;
					do {
						truncated_string = string.substring(0, string.length - remove_counter) + "…";
						string_width = ctx.measureText(truncated_string).width;
					}
					while ((string_width > max_width) && (++remove_counter < string.length));
					max_line_width = Math.max(max_line_width, string_width);
					current_line = truncated_string;
					break;
				}
			}
			lines.push(current_line);
		}
		lines.widest_line = max_line_width;
		return lines;
	}

	let defs, pattern_element, x_diagonal, y_diagonal, x_grid, y_grid, x_dots, y_dots;

	class Paths {
		constructor(s, r) {
			this.diagonal = `M 0 ${s} l ${s} -${s} M -${s * 0.25} ${s * 0.25} l ${s / 2} -${s/ 2} M ${s * 0.75} ${s * 1.25} l ${s / 2} -${s / 2}`;
			this.grid = `M ${s / 2} 0 l 0 ${s} M 0 ${s / 2} l ${s} 0`;
			this.dots = `M ${s / 4} ${s / 2} A ${r} ${r} 0 1 1 ${s / 4} ${(s / 2) + 0.1} Z`;
		}
	}

	const TILE_SIZE = 4;
	const DOT_RADIUS = 1.2;
	const PATHS = new Paths(TILE_SIZE, DOT_RADIUS);

	function initPatterns() {
		// these patterns were got from the textures npm package https://www.npmjs.com/package/textures

		pattern_element = create$1("svg:pattern")
			.attr("class", "fl-pattern-auto")
			.attr("width", TILE_SIZE)
			.attr("height", TILE_SIZE)
			.attr("patternUnits", "userSpaceOnUse");

		pattern_element
			.append("path")
			.attr("stroke-width", 1)
			.attr("shape-rendering", "auto")
			.attr("stroke-linecap", "square");

		x_diagonal = select(pattern_element.node().cloneNode(true)).attr("id", "fl-pattern-diagonal-x");
		x_diagonal.select("path").attr("d", PATHS.diagonal);

		x_grid = select(pattern_element.node().cloneNode(true)).attr("id", "fl-pattern-grid-x");
		x_grid.select("path").attr("d", PATHS.grid);

		x_dots = select(pattern_element.node().cloneNode(true)).attr("id", "fl-pattern-dots-x");
		x_dots.select("path").attr("d", PATHS.dots);

		y_diagonal = select(x_diagonal.node().cloneNode(true)).attr("id", "fl-pattern-diagonal-y");
		y_grid = select(x_grid.node().cloneNode(true)).attr("id", "fl-pattern-grid-y");
		y_dots= select(x_dots.node().cloneNode(true)).attr("id", "fl-pattern-dots-y");

		const patterns = [x_diagonal, x_dots, x_grid, y_diagonal, y_grid, y_dots];

		defs = create$1("svg:defs").node();

		patterns.forEach(p => {
			defs.appendChild(p.node());
		});
	}

	function initAppendPatternsTo(instance) {
		let add_patterns_to = instance.appendTo();
		initPatterns();

		instance.appendPatternsTo = function(container) {
			if (container === undefined) return add_patterns_to;
			container.node().appendChild(defs);
			add_patterns_to = container;
			return instance;
		};
	}

	function updateCustomPatterns(node, patterns) {
		const defs = node.select("defs");
		const customPattern = function() {
			return pattern_element.node().cloneNode(true);
		};

		const custom_pattern_select = defs.selectAll(".fl-pattern-custom").data(patterns);
		const custom_pattern_enter = custom_pattern_select.enter().append(customPattern);
		const custom_pattern_update = custom_pattern_select.merge(custom_pattern_enter);

		custom_pattern_update
			.attr("class", "fl-pattern-custom")
			.attr("id", (d, i) => `fl-pattern-custom-${i}`)
			.each(function(d) {
				const dots = d.texture === "fl-pattern-dots";
				const texture = d.texture.split("fl-pattern-")[1];
				select(this).select("path")
					.attr("d", PATHS[texture])
					.style("fill", dots ? d.color : "none")
					.style("stroke", dots ? "none" : d.color);
			});

		custom_pattern_select.exit().remove();
	}

	function updatePatterns(state) {
		x_dots.select("path").style("fill", state.x_area_color || "currentColor");
		x_diagonal.select("path").style("stroke", state.x_area_color || "currentColor");
		x_grid.select("path").style("stroke", state.x_area_color || "currentColor");
		y_dots.select("path").style("fill", state.y_area_color || "currentColor");
		y_diagonal.select("path").style("stroke", state.y_area_color || "currentColor");
		y_grid.select("path").style("stroke", state.y_area_color || "currentColor");
	}

	function initUpdate(instance, state) {
		// Cache the previous x and y scales so that entering lines are in correct position when there's also a scale change.
		// (See https://github.com/kiln/flourish-templates/issues/2313)
		// For now these are just used for lines (not areas)
		let previousXScale = {}, previousYScale = {};

		instance.update = function() {
			updatePatterns(state);
			const chart_layout = instance.chartLayout();
			const facet_name = instance.facetName();
			const offset = instance.offset();
			const line_height = 1.20;
			const data = [{ chart_layout, facet_name: facet_name}];
			const is_text_rtl = isTextRtl();

			const fg_label_class = "fg-label";
			const bg_label_class = "bg-label";

			const custom_patterns_set = new Set();

			var parseNum = instance.numberParser();
			var dur = instance.animationDuration();

			function getXTextSize() {
				return remToPx(state.x_font_size);
			}

			function getYTextSize() {
				return remToPx(state.y_font_size);
			}

			function splitLabelAndValue(string) {
				return string.replace(/\s*::\s*/g, "::").split("::").map(s => s.trim());
			}

			function getColors(highlight_color_str, state_highlight_color, state_text_color) {
				let colors = {
					highlight_color: state_highlight_color || "currentColor",
					text_color: state_text_color || "currentColor"
				};

				if (!highlight_color_str) return colors;

				if (highlight_color_str.includes("/")) {
					const split_colors = highlight_color_str.replace(/\s*\/\s*/g, "/").split("/").map(s => s.trim());
					colors.highlight_color = split_colors[0];
					colors.text_color = split_colors[1];
				}
				else {
					colors.highlight_color = highlight_color_str.trim();
					colors.text_color = state_text_color || "currentColor";
				}

				return colors;
			}

			function splitRange(string) {
				return string.replace(/\s*>>\s*/, ">>").split(">>").map(s => s.trim());
			}

			function getParsedValue(value, type, chart_layout, axis) {
				let parsed_value;
				if (type === "number") parsed_value = isNaN(parseNum(value)) ? parseNum(value) : value;
				else if (type === "datetime") {
					const parser = axis === "x" ? chart_layout.xDatetimeParse() : chart_layout.yDatetimeParse();
					parsed_value = parser(value);
				}
				else parsed_value = value;
				return parsed_value;
			}

			function valueIsWithinDomain(type, parsed_value, domain) {
				if (type === "number" || type === "datetime") {
					let sorted_domain = [...domain].sort((a, b) => a - b); // Pass in function so that dates are sorted correctly
					return ((parsed_value <= sorted_domain[1]) && (parsed_value >= sorted_domain[0]));
				}
				else {
					return domain.includes(parsed_value);
				}
			}

			function updateXLines(facet_group, data) {
				const chart_layout = data.chart_layout;
				let margin_left = chart_layout.margins().left;
				let margin_top = chart_layout.margins().top;
				let plot_height = Math.max(0, chart_layout.plot_height);
				let disabled = instance.xDisabled();
				let xScale = chart_layout.xScale();
				var enterXScale = previousXScale[facet_name] || xScale;
				let type = getTypeFromChartLayout(chart_layout.xData());

				var line_data = [];

				if (!disabled && state.x_enabled && state.x_lines) {
					state.x_lines.split("\n").forEach(function(string) {
						var line = splitLabelAndValue(string);
						var label = line[0];
						var value = line[1];
						if (value === undefined) return;
						var colors = getColors(line[2], state.x_line_color, state.x_label_color);
						var facet = line[3] || null;
						if (facet && data.facet_name !== facet) return;

						const parsed_value = getParsedValue(value, type, chart_layout, "x");
						const domain = xScale.domain();
						const value_is_valid = valueIsWithinDomain(type, parsed_value, domain);

						if (value_is_valid) {
							line_data.push({
								label: label,
								x: parsed_value,
								line_color: colors.highlight_color,
								label_color: colors.text_color
							});
						}
					});
				}

				var lines = select(facet_group).selectAll("g.line.x").data(line_data, d => d.label);

				var lines_enter = lines.enter().append("g").attr("class", "line x")
					.attr("pointer-events", "none")
					.attr("transform", function(d) { return "translate(" + (enterXScale(d.x) || margin_left) + ", " + margin_top + ")"; })
					.style("opacity", 0);
				lines_enter.append("line");
				lines_enter.append("text");

				var lines_update = lines.merge(lines_enter);
				lines_update.transition().duration(dur)
					.attr("transform", function(d) { return "translate(" + (xScale(d.x) || margin_left) + ", " + margin_top + ")"; })
					.style("opacity", 1);
				lines_update.select("line")
					.attr("stroke-dasharray", state.x_line_dash + " " + state.x_line_dash)
					.attr("stroke", d => d.line_color)
					.attr("stroke-width", state.x_line_width)
					.attr("y2", plot_height);
				lines_update.select("text").each(function(d) {
					const x_coordinate = (xScale(d.x) || margin_left);
					const font_family = instance.fontFamily();
					const max_width_in_px = remToPx(state.x_label_max_width);
					const font_style = `${state.x_font_weight}  ${state.x_font_size} ${font_family}`;
					const wrapped_lines_array = wrapStringToLines(d.label,
						font_style,
						null,
						max_width_in_px);

					drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, bg_label_class);
					drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, fg_label_class);

					const num_of_lines = wrapped_lines_array.length;
					const {transform_str, text_anchor} = getTransformAndAnchorForX(this, xScale, plot_height, num_of_lines, x_coordinate);
					select(this)
						.attr("transform", transform_str)
						.attr("text-anchor", text_anchor);
				})
				.attr("font-size", getXTextSize())
				.attr("fill", d => d.label_color);

				lines_update.selectAll(`.${bg_label_class}`)
					.attr("display", state.x_shadow_enabled ? "block": "none")
					.attr("fill", "none")
					.attr("stroke", state.x_shadow_color)
					.attr("stroke-width", 2);

				lines.exit()
					.transition().duration(dur)
					.attr("transform", function(d) { return "translate(" + (xScale(d.x) || margin_left) + ", " + margin_top + ")"; })
					.style("opacity", 0)
					.on("end", function() {
						select(this).remove();
					});
			}

			function getXAreaWidth(range, plot_width, margin_left) {
				const min_val = min(range);
				const max_val = max(range);
				const min_val_or_y_axis = Math.max(min_val, margin_left);
				const end_of_chart = plot_width + margin_left;
				const max_val_or_end_of_chart = Math.min(max_val, end_of_chart);
				const width = Math.abs(max_val_or_end_of_chart - min_val_or_y_axis);
				const backup_width = 4;
				// if width is 0 (because user has entered one invalid value) then just show highlight of width 4 around the valid value
				return width || backup_width;
			}

			function getCustomPatternString(color, texture) {
				return JSON.stringify({
					color: color,
					texture: texture
				});
			}

			function getXAreas(data) {
				const chart_layout = data.chart_layout;
				let margin_left = chart_layout.margins().left;
				let margin_top = chart_layout.margins().top;
				let plot_width = Math.max(0, chart_layout.plot_width);
				let xScale = chart_layout.xScale();
				let type = getTypeFromChartLayout(data.chart_layout.xData());
				let disabled = instance.xDisabled();
				const texture = state.x_area_texture;
				var area_data = [];

				if (!disabled && state.x_enabled && state.x_areas) {
					state.x_areas.split("\n").map(function(string) {
						var area = splitLabelAndValue(string);
						if (area.length < 2) return;
						var label = area[0];
						var split_range = splitRange(area[1]);
						var range = split_range.map(function(value, i) {
							var parsed_value = getParsedValue(value, type, chart_layout, "x");
							var x_val = parsed_value ? xScale(parsed_value) : null;
							if (instance.xExtendByStep()) x_val += chart_layout.xMinStep()/2 * (i == 0 ? -1 : 1);
							return x_val;
						});
						var colors = getColors(area[2], state.x_area_color, state.x_label_color);
						const color = colors.highlight_color;
						let custom_pattern_i = null;
						if (area[2] !== undefined && color !== "currentColor" && texture !== "none") {
							const pattern_string = getCustomPatternString(color, texture);
							custom_patterns_set.add(pattern_string);
							custom_pattern_i = [...custom_patterns_set].indexOf(pattern_string);
						}
						var facet = area[3] || null;
						if (facet && data.facet_name !== facet) return;

						const lower_parsed_value = getParsedValue(split_range?.[0] || "", type, chart_layout, "x");
						const upper_parsed_value = getParsedValue(split_range?.[1] || "", type, chart_layout, "x");
						const domain = xScale.domain();
						const lower_value_is_valid = valueIsWithinDomain(type, lower_parsed_value, domain);
						const upper_value_is_valid = valueIsWithinDomain(type, upper_parsed_value, domain);

						if (lower_value_is_valid || upper_value_is_valid) {
							if (range.length > 1) area_data.push({
								label: label,
								range: range,
								area_color: color,
								label_color: colors.text_color,
								transform: `translate(${Math.max(min(range), margin_left)}, ${margin_top})`,
								width: getXAreaWidth(range, plot_width, margin_left),
								height: Math.max(0, chart_layout.plot_height),
								x: (min(range) || margin_left),
								custom_pattern_i: custom_pattern_i
							});
						}
					});
				}
				return area_data;
			}

			function updateXAreas(facet_group, area_data, scale) {
				var areas = select(facet_group).selectAll("g.area.x").data(area_data, d => d.label);
				var areas_enter = areas.enter().append("g")
					.attr("class", "area x")
					.attr("pointer-events", "none")
					.attr("transform", d => d.transform)
					.style("opacity", 0);

				areas_enter.append("rect");
				areas_enter.append("text");

				var areas_update = areas.merge(areas_enter);

				areas_update.transition().duration(dur)
					.attr("transform", d => d.transform)
					.style("opacity", 1);

				areas_update.select("rect")
					.attr("width", d => d.width)
					.style("fill", d => getColorOrTexture(d, state.x_area_texture, "x"))
					.attr("opacity", state.x_area_opacity)
					.attr("height", d => d.height);

				areas_update.select("text").each(function(d) {
					const font_family = instance.fontFamily();
					const max_width_in_px = remToPx(state.x_label_max_width);
					const font_style = `${state.x_font_weight} ${state.x_font_size} ${font_family}`;

					const wrapped_lines_array = wrapStringToLines(d.label,
						font_style,
						null,
						max_width_in_px);
					drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, bg_label_class);
					drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, fg_label_class);

					const num_of_lines = wrapped_lines_array.length;
					const {transform_str, text_anchor} = getTransformAndAnchorForX(this, scale, d.height, num_of_lines, d.x);
					select(this)
						.attr("transform", transform_str)
						.attr("text-anchor", text_anchor);
				})
				.attr("font-size", getXTextSize())
				.attr("fill", d => d.label_color);

				areas_update.selectAll(`.${bg_label_class}`)
					.attr("display", state.x_shadow_enabled ? "block": "none")
					.attr("fill", "none")
					.attr("stroke", state.x_shadow_color || "transparent")
					.attr("stroke-width", 2);

				areas.exit()
					.transition().duration(dur)
					.style("opacity", 0)
					.on("end", function() {
						select(this).remove();
					});
			}

			function updateYLines(facet_group, data) {
				const chart_layout = data.chart_layout;
				let plot_width = Math.max(0, chart_layout.plot_width);
				let x = chart_layout.margins().left;
				let yScale = chart_layout.yScale();
				var enterYScale = previousYScale[facet_name] || yScale;
				let type = getTypeFromChartLayout(chart_layout.yData());
				let disabled = instance.yDisabled();
				let margin_top = chart_layout.margins().top;

				var line_data = [];
				if (!disabled && state.y_enabled && state.y_lines) {
					state.y_lines.split("\n").forEach(function(string) {
						var line = splitLabelAndValue(string);
						var label = line[0];
						var value = line[1];
						if (value === undefined) return;
						var colors = getColors(line[2], state.y_line_color, state.y_label_color);
						var facet = line[3] || null;
						if (facet && data.facet_name !== facet) return;

						const parsed_value = getParsedValue(value, type, chart_layout, "u");
						const domain = yScale.domain();
						const value_is_valid = valueIsWithinDomain(type, parsed_value, domain);

						if (value_is_valid) {
							line_data.push({
								label: label,
								y: parsed_value,
								line_color: colors.highlight_color,
								label_color: colors.text_color
							});
						}
					});
				}

				var lines = select(facet_group).selectAll("g.line.y").data(line_data, d => d.label);

				var lines_enter = lines.enter().append("g").attr("class", "line y")
					.attr("pointer-events", "none")
					.attr("transform", function(d) { return "translate(" + x + ", " + (enterYScale(d.y) || 0) + ")"; })
					.style("opacity", 0);
				lines_enter.append("line").attr("width", 0);
				lines_enter.append("text");

				var lines_update = lines.merge(lines_enter);
				lines_update.transition().duration(dur)
					.attr("transform", function(d) { return "translate(" + x + ", " + (yScale(d.y) || 0) + ")"; })
					.style("opacity", 1);
				lines_update.select("line")
					.attr("stroke-dasharray", state.y_line_dash + " " + state.y_line_dash)
					.attr("stroke", d => d.line_color)
					.attr("stroke-width", state.y_line_width)
					.attr("x2", plot_width);
				lines_update.select("text")
					.each(function(d) {
						const font_family = instance.fontFamily();
						const y_coordinate = (yScale(d.y) || margin_top);
						const max_width_in_px = remToPx(state.y_label_max_width);
						const font_style = `${state.y_font_weight} ${state.y_font_size} ${font_family}`;
						const wrapped_lines_array = wrapStringToLines(d.label,
							font_style,
							null,
							max_width_in_px);

						drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, bg_label_class);
						drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, fg_label_class);

						const num_of_lines = wrapped_lines_array.length;
						const {transform_str, text_anchor} = getTransformAndAnchorForY(this, yScale, plot_width, num_of_lines, y_coordinate);

						select(this)
							.attr("transform", transform_str)
							.attr("text-anchor", text_anchor);
					})
					.attr("font-size", getYTextSize())
					.attr("fill", d => d.label_color);

				lines_update.selectAll(`.${bg_label_class}`)
					.attr("display", state.y_shadow_enabled ? "block": "none")
					.attr("fill", "none")
					.attr("stroke", state.y_shadow_color)
					.attr("stroke-width", 2);

				lines.exit()
					.transition().duration(dur)
					.attr("transform", function(d) { return "translate(" + x + ", " + (yScale(d.y) || 0) + ")"; })
					.style("opacity", 0)
					.on("end", function() {
						select(this).remove();
					});
			}

			function getYAreaHeight(range, plot_height, margin_top) {
				const min_val = min(range);
				const max_val = max(range);
				const bottom_of_chart = plot_height + margin_top;
				const max_val_or_bottom_of_chart = Math.min(max_val, bottom_of_chart);
				const min_val_or_top_of_chart = Math.max(min_val, margin_top);
				const height = Math.abs(max_val_or_bottom_of_chart - min_val_or_top_of_chart);
				const backup_height = 4;
				// if height is 0 (because user has entered one invalid value) then just show highlight of height 4 around the valid value
				return height || backup_height;
			}

			function getYAreas(data) {
				const chart_layout = data.chart_layout;
				let margin_top = chart_layout.margins().top;
				let plot_height = Math.max(0, chart_layout.plot_height);
				let margin_left = chart_layout.margins().left;
				let yScale = chart_layout.yScale();
				let type = getTypeFromChartLayout(chart_layout.yData());
				let disabled = instance.yDisabled();
				const texture = state.y_area_texture;
				var area_data = [];

				if (!disabled && state.y_enabled && state.y_areas) {
					state.y_areas.split("\n").map(function(string) {
						var area = splitLabelAndValue(string);
						if (area.length < 2) return;
						var label = area[0];
						var split_range = splitRange(area[1]);
						var range = split_range.map(function(value, i) {
							var parsed_value = getParsedValue(value, type, chart_layout, "y");
							var y_val = parsed_value ? yScale(parsed_value) : parsed_value;
							if (instance.yExtendByStep()) y_val += chart_layout.yMinStep()/2 * (i == 0 ? 1 : -1);
							return y_val;
						});
						var colors = getColors(area[2], state.y_area_color, state.y_label_color);
						const color = colors.highlight_color;
						let custom_pattern_i = null;
						if (area[2] !== undefined && color !== "currentColor" && texture !== "none") {
							const pattern_string = getCustomPatternString(color, texture);
							custom_patterns_set.add(pattern_string);
							custom_pattern_i = [...custom_patterns_set].indexOf(pattern_string);
						}
						var facet = area[3] || null;
						if (facet && data.facet_name !== facet) return;

						const lower_parsed_value = getParsedValue(split_range?.[0] || "", type, chart_layout, "y");
						const upper_parsed_value = getParsedValue(split_range?.[1] || "", type, chart_layout, "y");
						const domain = yScale.domain();
						const lower_value_is_valid = valueIsWithinDomain(type, lower_parsed_value, domain);
						const upper_value_is_valid = valueIsWithinDomain(type, upper_parsed_value, domain);

						if (lower_value_is_valid || upper_value_is_valid) {
							if (range.length > 1) area_data.push({
								label: label,
								range: range,
								area_color: color,
								label_color: colors.text_color,
								width: Math.max(0, chart_layout.plot_width),
								height: getYAreaHeight(range, plot_height, margin_top),
								transform: `translate(${margin_left}, ${Math.max(min(range), margin_top)})`,
								y: (min(range) || margin_top),
								custom_pattern_i: custom_pattern_i
							});
						}
					});
				}
				return area_data;
			}

			function updateYAreas(facet_group, area_data, scale) {
				var areas = select(facet_group).selectAll("g.area.y").data(area_data, d => d.label);

				var areas_enter = areas.enter().append("g").attr("class", "area y")
					.attr("pointer-events", "none")
					.attr("transform", d => d.transform)
					.style("opacity", 0);

				areas_enter.append("rect");
				areas_enter.append("text");

				var areas_update = areas.merge(areas_enter);

				areas_update
					.transition().duration(dur)
					.attr("transform", d => d.transform)
					.style("opacity", 1);

				areas_update.select("rect")
					.attr("height", d => d.height)
					.attr("fill", d => getColorOrTexture(d, state.y_area_texture, "y"))
					.attr("opacity", state.y_area_opacity)
					.attr("width", d => d.width);

				areas_update.select("text").each(function(d) {
					const font_family = instance.fontFamily();
					const max_width_in_px = remToPx(state.y_label_max_width);
					const font_style = `${state.y_font_weight} ${state.y_font_size} ${font_family}`;

					const wrapped_lines_array = wrapStringToLines(d.label,
						font_style,
						null,
						max_width_in_px);

					drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, bg_label_class);
					drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, fg_label_class);
					const num_of_lines = wrapped_lines_array.length;
					const { transform_str, text_anchor } = getTransformAndAnchorForY(this, scale, d.width, num_of_lines, d.y);

					select(this)
						.attr("transform", transform_str)
						.attr("text-anchor", text_anchor);
				})
				.attr("font-size", getYTextSize())
				.attr("fill", d => d.label_color);

				areas_update.selectAll(`.${bg_label_class}`)
					.attr("display", state.y_shadow_enabled ? "block": "none")
					.attr("fill", "none")
					.attr("stroke", state.y_shadow_color || "transparent")
					.attr("stroke-width", 2);

				areas.exit()
					.transition().duration(dur)
					.style("opacity", 0)
					.on("end", function() {
						select(this).remove();
					});
			}

			function drawWrappedLabels(wrapped_lines_array, target_el, font_size, font_weight, class_name) {
				var font_size_in_px = remToPx(font_size);
				var tspans = select(target_el).selectAll(`.${class_name}`).data(wrapped_lines_array);

				tspans.join("tspan")
						.text(function (d) { return d ; })
						.attr("class", class_name)
						.attr("y", function(_, i) {
							var y = (i * (font_size_in_px * line_height));
							return y + "px";
						})
						.attr("x", 0)
						.attr("font-weight", font_weight);

				tspans.exit().remove();
			}

			function getTransformAndAnchorForX(target_el, xScale, chart_height, num_of_lines, x_coordinate_of_highlight) {
				/* the reason we are working out the text anchor and the transform at
				the same time is because we only want to calculate if a label overlaps
				the chart once, and that affects both properties*/
				let transform_str;
				let x_position_of_text = 0;
				let text_anchor;

				const x_position_middle = state.x_label_x_position == "middle";
				const y_position_bottom = state.x_label_y_position == "bottom";
				const is_vertical = state.x_label_vertical;
				const line_height_in_px = getXTextSize() * line_height;
				// svg draws from the bottom of the first line of text, so we need
				// to times line height by num of lines minus one
				const text_block_offset = line_height_in_px * (num_of_lines - 1);

				const x_padding = getXTextSize() /2;
				const y_padding_if_horizontal = y_position_bottom ? text_block_offset + x_padding: line_height_in_px;
				const y_padding_if_vertical = x_padding;
				const y_padding = is_vertical ? y_padding_if_vertical : y_padding_if_horizontal;
				const vertical_center =  is_vertical ? (chart_height / 2) : (chart_height / 2) - (text_block_offset / 2);
				const max_x_value = xScale.range()[1];
				const min_x_value = xScale.range()[0];
				const text_width = target_el.getBoundingClientRect().width;

				const y_position_lookup = {
					top: y_padding,
					middle: vertical_center,
					bottom: chart_height - y_padding
				};
				const y_position = y_position_lookup[state.x_label_y_position];

				if (is_vertical) {
					const x_position_lookup = {
						right: getXTextSize(),
						middle: - (text_block_offset - x_padding)/2,
						left: - text_block_offset - x_padding
					};
					x_position_of_text = x_position_lookup[state.x_label_x_position];

					const text_anchor_lookup = {
						top: is_text_rtl ? "start" : "end",
						middle: "middle",
						bottom: is_text_rtl ? "end" : "start"
					};
					text_anchor = text_anchor_lookup[state.x_label_y_position];

					const overlap_at_end = x_coordinate_of_highlight + text_width + x_padding > max_x_value;

					const overlap_at_start = x_coordinate_of_highlight - text_width - x_padding < min_x_value;

					if (overlap_at_end) x_position_of_text = x_position_lookup.left;
					if (overlap_at_start) x_position_of_text = x_position_lookup.right;

					if (x_position_middle) {
						const overlap_at_end = x_coordinate_of_highlight + (text_width /2) > max_x_value;
						const overlap_at_start = x_coordinate_of_highlight - (text_width/2) < min_x_value;

						if (overlap_at_end) x_position_of_text = x_position_lookup.left;
						if (overlap_at_start) x_position_of_text = x_position_lookup.right;
					}

					transform_str = "translate(" + x_position_of_text + ", " + y_position + ") rotate(-90)";
				}
				else if (!is_vertical) {
					// Add padding depending on which side of the line the text is on
					// If there is not enough space for the label, it will move it to
					// the other side.

					const x_position_lookup = {
						right: x_padding,
						middle: 0,
						left: - x_padding
					};
					x_position_of_text = x_position_lookup[state.x_label_x_position];

					const text_anchor_lookup = {
						right: is_text_rtl ? "end": "start",
						middle: "middle",
						left: is_text_rtl ? "start" : "end"
					};
					text_anchor = text_anchor_lookup[state.x_label_x_position];

					let overlap_at_end;
					let overlap_at_start;

					if (x_position_middle) {
						overlap_at_end = x_coordinate_of_highlight + (text_width /2) > max_x_value;
						overlap_at_start = x_coordinate_of_highlight - (text_width/2) < min_x_value;
					}
					else {
						overlap_at_end = x_coordinate_of_highlight + text_width + x_padding > max_x_value;
						overlap_at_start = x_coordinate_of_highlight - text_width - x_padding < min_x_value;
					}

					if (overlap_at_end) {
						x_position_of_text = x_position_lookup.left;
						text_anchor = text_anchor_lookup.left;
					}
					if (overlap_at_start) {
						x_position_of_text = x_position_lookup.right;
						text_anchor = text_anchor_lookup.right;
					}

					transform_str = "translate(" + x_position_of_text + ", " + y_position + ")";
				}

				return {
					transform_str,
					text_anchor
				};
			}

			function getTransformAndAnchorForY(target_el, yScale, chart_width, num_of_lines, y_coordinate_of_line) {
				let text_anchor;
				let transform_str;

				const text_anchor_lookup = {
					left: is_text_rtl ? "end" : "start",
					middle: "middle",
					right: is_text_rtl ? "start": "end"
				};
				text_anchor = text_anchor_lookup[state.y_label_x_position];

				const line_height_in_px = getYTextSize() * line_height;
				const text_block_offset = line_height_in_px * (num_of_lines - 1);
				const x_padding = getYTextSize()/2;
				const text_offset_and_padding = text_block_offset + x_padding;
				const max_y_val = yScale.range()[0];
				const min_y_val = yScale.range()[1];
				const text_height = target_el.getBoundingClientRect().height;
				const y_position_middle = state.y_label_y_position == "middle";

				const y_position_lookup = {
					above: - text_offset_and_padding,
					middle: - (text_offset_and_padding/2) + x_padding,
					below: line_height_in_px
				};
				let y_position = y_position_lookup[state.y_label_y_position];

				const x_position_lookup = {
					right: chart_width - x_padding,
					middle: chart_width / 2,
					left: x_padding,
				};
				let x_position = x_position_lookup[state.y_label_x_position];

				let overlap_at_top;
				let overlap_at_bottom;

				if (y_position_middle) {
					overlap_at_top = y_coordinate_of_line - (text_height/2) < min_y_val;
					overlap_at_bottom = y_coordinate_of_line + (text_height/2) > max_y_val;
				}
				else {
					overlap_at_top = y_coordinate_of_line - text_height < min_y_val;
					overlap_at_bottom = y_coordinate_of_line + text_height > max_y_val;
				}

				if (overlap_at_top) y_position = y_position_lookup.below;
				if (overlap_at_bottom) y_position = y_position_lookup.above;

				transform_str = "translate(" + x_position + ", " + y_position + ")";

				return {
					transform_str,
					text_anchor
				};
			}

			function updateHighlights() {
				var parent_element = instance.appendTo();
				var axis_highlights = parent_element.selectAll("g.axis-highlights").data(data);
				var axis_highlights_enter = axis_highlights.enter().append("g").attr("class", "axis-highlights")
					.style("opacity", 0);
				axis_highlights.merge(axis_highlights_enter)
					.transition().duration(dur)
					.style("opacity", 1)
					.attr("transform", function() {
						if (!offset) return null;
						if (isNaN(offset.x) || isNaN(offset.y)) return null;
						return "translate(" + offset.x + ", " + offset.y + ")";
					})
					.each(function(d) {
						var facet_group = this;
						const x_areas = getXAreas(d);
						const y_areas = getYAreas(d);
						const custom_patterns = [...custom_patterns_set].map(c => JSON.parse(c));
						updateCustomPatterns(parent_element, custom_patterns);
						updateXAreas(facet_group, x_areas, d.chart_layout.xScale());
						updateYAreas(facet_group, y_areas, d.chart_layout.yScale());
						updateXLines(facet_group, d);
						updateYLines(facet_group, d);
					});

				axis_highlights.exit().remove();

				// Cache scale for next update
				previousXScale[facet_name] = chart_layout.xScale();
				previousYScale[facet_name] = chart_layout.yScale();
			}

			updateHighlights();
		};
	}

	function initFacetName(instance) {
		let facet_name;

		instance.facetName = function(value) {
			if (value === undefined) return facet_name;
			facet_name = value;
			return instance;
		};
	}

	function initAppendTo(instance) {
		let append_to;

		instance.appendTo = function(value) {
			if (value === undefined) return append_to;
			// make the patterns default to being appended to the appendTo container.
			append_to = value;
			if (instance.appendPatternsTo() === undefined) {
				instance.appendPatternsTo(append_to);
			}
			return instance;
		};
	}

	function initFontFamily(instance) {
		let font_family;

		instance.fontFamily = function(value) {
			if (value === undefined) return font_family;
			font_family = value;
			return instance;
		};
	}

	function initOffset(instance) {
		let offset;

		instance.offset = function (value) {
			if (value === undefined) return offset;
			offset = value;
			return instance;
		};
	}

	function init(state) {
		for (const key in DEFAULT_STATE) {
			if (state[key] === undefined) state[key] = DEFAULT_STATE[key];
		}

		const instance = {};
		initAnimationDuration(instance);
		initNumberParser(instance);
		initAppendTo(instance);
		initChartLayout(instance);
		initXDisabled(instance);
		initYDisabled(instance);
		initXExtendByStep(instance);
		initYExtendByStep(instance);
		initFacetName(instance);
		initFontFamily(instance);
		initAppendPatternsTo(instance);
		initOffset(instance);
		initUpdate(instance, state);

		return instance;
	}

	return init;

}));
