import {create, select} from "d3-selection";

let defs, pattern_element, x_diagonal, y_diagonal, x_grid, y_grid, x_dots, y_dots;

class Paths {
	constructor(s, r) {
		this.diagonal = `M 0 ${s} l ${s} -${s} M -${s * 0.25} ${s * 0.25} l ${s / 2} -${s/ 2} M ${s * 0.75} ${s * 1.25} l ${s / 2} -${s / 2}`;
		this.grid = `M ${s / 2} 0 l 0 ${s} M 0 ${s / 2} l ${s} 0`;
		this.dots = `M ${s / 4} ${s / 2} A ${r} ${r} 0 1 1 ${s / 4} ${(s / 2) + 0.1} Z`;
	}
}

const TILE_SIZE = 4;
const DOT_RADIUS = 1.2;
const PATHS = new Paths(TILE_SIZE, DOT_RADIUS);

function initPatterns() {
	// these patterns were got from the textures npm package https://www.npmjs.com/package/textures

	pattern_element = create("svg:pattern")
		.attr("class", "fl-pattern-auto")
		.attr("width", TILE_SIZE)
		.attr("height", TILE_SIZE)
		.attr("patternUnits", "userSpaceOnUse");

	pattern_element
		.append("path")
		.attr("stroke-width", 1)
		.attr("shape-rendering", "auto")
		.attr("stroke-linecap", "square");

	x_diagonal = select(pattern_element.node().cloneNode(true)).attr("id", "fl-pattern-diagonal-x");
	x_diagonal.select("path").attr("d", PATHS.diagonal);

	x_grid = select(pattern_element.node().cloneNode(true)).attr("id", "fl-pattern-grid-x");
	x_grid.select("path").attr("d", PATHS.grid);

	x_dots = select(pattern_element.node().cloneNode(true)).attr("id", "fl-pattern-dots-x");
	x_dots.select("path").attr("d", PATHS.dots);

	y_diagonal = select(x_diagonal.node().cloneNode(true)).attr("id", "fl-pattern-diagonal-y");
	y_grid = select(x_grid.node().cloneNode(true)).attr("id", "fl-pattern-grid-y");
	y_dots= select(x_dots.node().cloneNode(true)).attr("id", "fl-pattern-dots-y");

	const patterns = [x_diagonal, x_dots, x_grid, y_diagonal, y_grid, y_dots];

	defs = create("svg:defs").node();

	patterns.forEach(p => {
		defs.appendChild(p.node());
	});
}

export function initAppendPatternsTo(instance) {
	let add_patterns_to = instance.appendTo();
	initPatterns();

	instance.appendPatternsTo = function(container) {
		if (container === undefined) return add_patterns_to;
		container.node().appendChild(defs);
		add_patterns_to = container;
		return instance;
	};
}

export function updateCustomPatterns(node, patterns) {
	const defs = node.select("defs");
	const customPattern = function() {
		return pattern_element.node().cloneNode(true);
	};

	const custom_pattern_select = defs.selectAll(".fl-pattern-custom").data(patterns);
	const custom_pattern_enter = custom_pattern_select.enter().append(customPattern);
	const custom_pattern_update = custom_pattern_select.merge(custom_pattern_enter);

	custom_pattern_update
		.attr("class", "fl-pattern-custom")
		.attr("id", (d, i) => `fl-pattern-custom-${i}`)
		.each(function(d) {
			const dots = d.texture === "fl-pattern-dots";
			const texture = d.texture.split("fl-pattern-")[1];
			select(this).select("path")
				.attr("d", PATHS[texture])
				.style("fill", dots ? d.color : "none")
				.style("stroke", dots ? "none" : d.color);
		});

	custom_pattern_select.exit().remove();
}

export function updatePatterns(state) {
	x_dots.select("path").style("fill", state.x_area_color || "currentColor");
	x_diagonal.select("path").style("stroke", state.x_area_color || "currentColor");
	x_grid.select("path").style("stroke", state.x_area_color || "currentColor");
	y_dots.select("path").style("fill", state.y_area_color || "currentColor");
	y_diagonal.select("path").style("stroke", state.y_area_color || "currentColor");
	y_grid.select("path").style("stroke", state.y_area_color || "currentColor");
}
